

from dataclasses import dataclass

from ..configfiles.Configfile import Configfile
from .ToolXMLMetadata import ToolXMLMetadata
from .param.ParamRegister import ParamRegister
from .TestRegister import TestRegister
from .requirements import ContainerRequirement, CondaRequirement

Requirement = ContainerRequirement | CondaRequirement


@dataclass
class XMLToolDefinition:
    """
    High-level component representing a tool. 
    Does not depend on lower level representations or parsing.
    Permits storing and retreiving data about the tool.
    """
    metadata: ToolXMLMetadata
    raw_command: str
    configfiles: list[Configfile]
    inputs: ParamRegister
    outputs: ParamRegister
    tests: TestRegister

    # def get_input(self, query: str, strategy: str='exact') -> Optional[Param]:
    #     return self.inputs.get(query.lstrip('$'), strategy=strategy)
    
    # def list_inputs(self) -> list[Param]:
    #     return self.inputs.list()

    # def get_output(self, query: str, strategy: str='exact') -> Optional[Param]:
    #     return self.outputs.get(query.lstrip('$'), strategy=strategy)
    
    # def list_outputs(self) -> list[OutputParam]:
    #     return self.outputs.list()

    # def list_tests(self) -> list[TTestCase]:
    #     return self.tests.list()

    # def get_requirements(self) -> list[Requirement]:
    #     return self.metadata.requirements
    
    # def get_main_requirement(self) -> Requirement:
    #     return self.metadata.get_main_requirement()


















from abc import ABC, abstractmethod
from dataclasses import dataclass


"""
EXAMPLE
<requirements>
    <requirement type="package" version="1.2.0">bowtie</requirement>
    <container type="docker">google/deepvariant:@TOOL_VERSION@</container>
</requirements>
"""


@dataclass
class Requirement(ABC):
    """models a tool XML requirement"""

    @property
    @abstractmethod
    def name(self) -> str:
        ...
    
    @property
    @abstractmethod
    def version(self) -> str:
        ...
    
    # @abstractmethod
    # def to_dict(self) -> dict[str, str]:
    #     ...



@dataclass
class CondaRequirement(Requirement):
    _name: str
    _version: str
    subtype: str = 'conda'

    @property
    def name(self) -> str:
        return self._name

    @property
    def version(self) -> str:
        return self._version
    
    # def to_dict(self) -> dict[str, str]:
    #     return {
    #         'type': self.subtype,
    #         'name': self.name,
    #         'version': self.version
    #     }


@dataclass
class ContainerRequirement(Requirement):
    url: str
    flavour: str
    registry_host: str
    image_type: str
    subtype: str = 'container'

    @property
    def name(self) -> str:
        return self.url.rsplit('/', 1)[-1].split(':', 1)[0]
    
    @property
    def version(self) -> str:
        return self.url.rsplit('/', 1)[-1].split(':', 1)[-1]

    # def to_dict(self) -> dict[str, str]:
    #     return {
    #         'type': self.subtype,
    #         'url': self._url,
    #         'flavour': self._flavour,
    #         'registry_host': self._registry_host
    #     }








from galaxy2janis.logs import logging
from galaxy2janis import settings
import json

from typing import Any, Optional
from galaxy2janis.startup import workflow_setup

from galaxy2janis.entities.workflow import Workflow

from galaxy2janis.gx.gxworkflow.parsing.metadata import ingest_metadata
from galaxy2janis.gx.gxworkflow.parsing.inputs import ingest_workflow_inputs
from galaxy2janis.gx.gxworkflow.parsing.step import ingest_workflow_steps
from galaxy2janis.gx.gxworkflow.parsing.tool_step.tool import ingest_workflow_tools
from galaxy2janis.gx.gxworkflow.parsing.tool_step.prepost import ingest_workflow_steps_prepost
from galaxy2janis.gx.gxworkflow.parsing.tool_step.outputs import ingest_workflow_steps_outputs

from galaxy2janis.gx.gxworkflow.values import handle_tool_connection_inputs
from galaxy2janis.gx.gxworkflow.values import handle_tool_runtime_inputs
from galaxy2janis.gx.gxworkflow.values import handle_tool_static_inputs
from galaxy2janis.gx.gxworkflow.values import handle_tool_default_inputs

from galaxy2janis.gx.gxworkflow.updates import update_component_knowledge
from galaxy2janis.gx.gxworkflow.connections import handle_scattering
from galaxy2janis.gx.gxworkflow.values.scripts import handle_tool_script_inputs

from galaxy2janis.fileio import write_workflow



"""
this file is the overall orchestrator to parse 
a galaxy workflow to janis. 
class approach used just to reduce number of variables
being passed to functions here (clutter) and to reduce
the number of times certain files are loaded (speed)
the order here seems weird but trust me there is reason. 
"""

def workflow_mode(args: dict[str, Optional[str]]) -> None:
    workflow_setup(args)
    logging.msg_parsing_workflow()

    galaxy = load_tree()
    janis = Workflow()

    # ingesting workflow entities to internal
    ingest_metadata(janis, galaxy)
    ingest_workflow_inputs(janis, galaxy)
    ingest_workflow_steps(janis, galaxy)
    ingest_workflow_tools(janis, galaxy)
    ingest_workflow_steps_prepost(janis, galaxy)
    ingest_workflow_steps_outputs(janis, galaxy) 

    # assigning tool input values
    handle_tool_connection_inputs(janis, galaxy)
    handle_tool_runtime_inputs(janis, galaxy)
    handle_tool_script_inputs(janis)
    handle_tool_static_inputs(janis, galaxy)
    handle_tool_default_inputs(janis)

    update_component_knowledge(janis)
    handle_scattering(janis)
    write_workflow(janis)

def load_tree() -> dict[str, Any]:
    # TODO should probably check the workflow type (.ga, .ga2)
    # and internal format is valid
    with open(settings.workflow.workflow_path, 'r') as fp:
        return json.load(fp)







import regex as re

from .patterns import (
    INTEGER,
    FLOAT,
    SCRIPT,
    VARIABLES_FMT1,
    VARIABLES_FMT2,
)

from .matches import get_matches

def is_int(the_string: str) -> bool:
    matches = get_matches(the_string, INTEGER)
    if matches:
        return True
    return False

def is_float(the_string: str) -> bool:
    matches = get_matches(the_string, FLOAT)
    if matches:
        return True
    return False

def is_var(the_string: str) -> bool:
    matches1 = get_matches(the_string, VARIABLES_FMT1)
    matches2 = get_matches(the_string, VARIABLES_FMT2)
    if matches1 or matches2:
        return True
    return False

def is_script(the_string: str) -> bool:
    matches = get_matches(the_string, SCRIPT)
    if matches:
        return True
    return False

def has_var(the_string: str) -> bool:
    if '$' in the_string:
        return True
    return False

def is_present(word: str, text: str) -> bool:
    pattern = rf'\s{word}\s'
    if re.findall(pattern, text):
        return True
    return False




from __future__ import annotations
from typing import Any, Optional

from ....gxtool.param.Param import Param
from ..ValueRecord import ValueRecord
from .InputComponent import InputComponent
from . import utils


class Option(InputComponent):
    def __init__(self, prefix: str) -> None:
        super().__init__()
        self.prefix = prefix
        self.delim: str = ' '
        self.gxparam_attachment: int = 1
        self.values: ValueRecord = ValueRecord()

    @property
    def name(self) -> str:
        return self.prefix.strip('--')

    @property
    def default_value(self) -> Any:
        """gets the default value for this component"""
        if self.gxparam:
            default = self.gxparam.default
        elif len(self.values.unique) == 1:
            default = self.values.unique[0]
        elif len(self.values.unique) > 1:
            if self.values.script:
                default = self.values.script
            elif self.values.env_var:
                default = self.values.env_var
            else:
                default = self.values.most_common_value
        else:
            default = None
        return utils.sanitise_default_value(default)
    
    @property
    def optional(self) -> bool:
        if self.forced_optionality is not None:
            return self.forced_optionality
        elif self.gxparam:
            return self.gxparam.optional
        return False

    @property
    def array(self) -> bool:
        if self.forced_array is not None:
            return self.forced_array
        elif self.gxparam:
            return self.gxparam.array
        return False

    @property
    def docstring(self) -> Optional[str]:
        if self.gxparam:
            return self.gxparam.docstring
        return ''
        #return f'examples: {", ".join(self.values.unique[:3])}'

    def update(self, incoming: Any):
        assert(isinstance(incoming, Option))
        # transfer values
        self.values.record += incoming.values.record
        # transfer galaxy param reference
        if not self.gxparam and incoming.gxparam:
            self.gxparam: Optional[Param] = incoming.gxparam

    def __str__(self) -> str:
        return f'{str(self.prefix):30}{str(self.default_value):20}{str(self.optional):>10}'








"""
Module governs tags within a workflow. 
Only the entities in this file should be accessed from external locations. 
"""

from .manager import get
from .manager import register
from .manager import new_group
from .manager import switch_group





from dataclasses import dataclass
from typing import Any, Optional
from uuid import uuid4

from galaxy2janis.datatypes import JanisDatatype
from galaxy2janis import tags


@dataclass
class WorkflowInput:
    _name: str
    array: bool
    optional: bool
    is_runtime: bool
    datatype: JanisDatatype
    value: Any = None

    def __post_init__(self):
        self.uuid: str = str(uuid4())

    @property
    def name(self) -> str:
        if self.is_runtime:
            return self._name
        else:
            return f'in_{self._name}'
    
    @property
    def tag(self) -> str:
        return tags.get(self.uuid)
    
    @property
    def docstring(self) -> Optional[str]:
        return 'None yet!'
        






from collections import defaultdict
from typing import Optional

from galaxy2janis import expressions


class ValueRecord:
    def __init__(self):
        self.record: list[str] = []

    def add(self, value: str) -> None:
        self.record.append(value)
    
    @property
    def unique(self) -> list[str]:
        values = list(set([text for text in self.record]))
        values.sort()
        return values
    
    @property
    def counts(self) -> defaultdict[str, int]:
        counts: defaultdict[str, int] = defaultdict(int) 
        for text in self.record:
            if text != '': # TODO how would this happen ??
                counts[text] += 1
        return counts

    @property
    def most_common_value(self) -> Optional[str]:
        counts_dict = self.counts
        counts_list = list(counts_dict.items())
        counts_list.sort(key=lambda x: x[1], reverse=True)
        if len(counts_list) > 0:
            return counts_list[0][0]
        else:
            return None

    @property
    def env_var(self) -> Optional[str]:
        for text in self.record:
            if text.startswith('$'):
                return text
        return None
    
    @property
    def script(self) -> Optional[str]:
        for text in self.record:
            if expressions.is_script(text):
                return text
        return None

    










from typing import Any
from .groups import TagGroup

groups: dict[str, TagGroup] = {}


def register(entity: Any) -> None:  
    """register a tag in the active TagGroup"""
    group = _get_active()
    group.register(entity.name, entity) 

def get(uuid: str) -> str:
    """get a tag from any TagGroup"""
    global groups
    for group in groups.values():
        tag = group.get(uuid)
        if tag:
            return tag
    raise RuntimeError('No tag registered')

def new_group(section: str, uuid: str):
    """
    create new TagGroup (ie for a new workflow or tool), 
    then switch to that TagGroup to start registering tags
    """
    global groups
    groups[uuid] = TagGroup(section)
    switch_group(uuid)

def switch_group(uuid: str):
    """swap to a specific TagGroup to register new tags in that group"""
    _clear_active()
    _set_active(uuid)
            

def _get_active() -> TagGroup:
    global groups
    for group in groups.values():
        if group.active:
            return group
    raise RuntimeError('no active group')

def _set_active(query_uuid: str):
    global groups
    for uuid, group in groups.items():
        if uuid == query_uuid:
            group.active = True
            return
    raise RuntimeError('group doesnt exist')

def _clear_active():
    global groups
    for group in groups.values():
        group.active = False











from Bio import pairwise2


def global_align(pattern: str, template: str) -> int:
    pattern = pattern.lower()
    template = template.lower()
    outcome = pairwise2.align.globalms(pattern, template, 2, -1, -.5, -.1) # type: ignore
    if len(outcome) > 0: # type: ignore
        score = outcome[0].score # type: ignore
    else:
        score = 0
    return score # type: ignore







from typing import Any

value_translations = {
        'false': False,
        'true': True,
    }


def standardise_tool_state(gxstep: dict[str, Any]) -> dict[str, Any]:
    out: dict[str, Any] = {}
    for key, val in gxstep['tool_state'].items():
        out[key] = standardise_tool_state_value(val)
    return out

def standardise_tool_state_value(value: Any) -> Any:
    value = handle_array(value)
    value = handle_translations(value)
    return value

def handle_array(value: Any) -> Any:
    if isinstance(value, list):
        return None
        # if len(value) == 0:
        #     value = None
        # else:
        #     value = value[0]
    return value

def handle_translations(value: Any) -> Any:
    if value in value_translations:
        return value_translations[value]
    return value






from galaxy2janis.logs import logging
from galaxy2janis import expressions

from typing import Optional

from galaxy2janis.gx.gxtool import XMLToolDefinition
from galaxy2janis.expressions.patterns import LINUX_STATEMENT_DELIMS

from .CommandString import CommandString
from .DynamicCommandStatement import DynamicCommandStatement
from .MainStatementInferrer import MainStatementInferrer
from .RealisedTokenValues import RealisedTokenFactory


def gen_command_string(
    source: str, 
    the_string: str, 
    xmltool: Optional[XMLToolDefinition]=None,
    requirement: Optional[str]=None
    ) -> CommandString:

    if xmltool and not requirement:
        requirement = xmltool.metadata.get_main_requirement().name
    assert(requirement)

    statements = _gen_command_statements(the_string, xmltool)
    statement_dict = _split_pre_main_post_statements(statements, source, requirement)
    return _init_command_string(statement_dict)

def gen_command_statement(statement: str, xmltool: Optional[XMLToolDefinition]=None) -> DynamicCommandStatement:
    factory = RealisedTokenFactory(xmltool)
    realised_tokens = factory.try_tokenify(statement)
    return DynamicCommandStatement(statement, realised_tokens)

def _gen_command_statements(the_string: str, xmltool: Optional[XMLToolDefinition]=None) -> list[DynamicCommandStatement]:
    statements: list[DynamicCommandStatement] = []
    for cmdstmt in _split_text_statements(the_string):
        statements.append(gen_command_statement(cmdstmt, xmltool))
    return statements

def _split_text_statements(the_string: str) -> list[str]:
    statements: list[str] = []
    
    delim_matches = expressions.get_matches(the_string, LINUX_STATEMENT_DELIMS)
    quoted_sections = expressions.get_quoted_sections(the_string)

    # has to be reverse order otherwise m.start() and m.end() are out of place
    for m in sorted(delim_matches, key=lambda x: x.start(), reverse=True): 
        if quoted_sections[m.start()] == False and quoted_sections[m.end()] == False:
            left_split = the_string[:m.start()]
            right_split = the_string[m.end():]
            statements = [right_split] + statements # prepend
            the_string = left_split

    statements = [the_string] + statements # prepend the final split (left this time)
    return statements

def _split_pre_main_post_statements(
    statements: list[DynamicCommandStatement], 
    source: str, 
    requirement: str) -> dict[str, list[DynamicCommandStatement]]:

    out: dict[str, list[DynamicCommandStatement]] = {'pre': [], 'main': [], 'post': []}
    inferrer = MainStatementInferrer(statements, source, requirement)
    main_index = inferrer.infer()
    
    for i, statement in enumerate(statements):
        if i < main_index:
            out['pre'].append(statement)
        elif i == main_index:
            out['main'].append(statement)
        else:
            out['post'].append(statement)

    return out

def _init_command_string(statement_dict: dict[str, list[DynamicCommandStatement]]) -> CommandString:
    if statement_dict['pre']:
        logging.has_preprocessing()
    if statement_dict['post']:
        logging.has_postprocessing()
    return CommandString(
        main=statement_dict['main'][0], 
        preprocessing=statement_dict['pre'], 
        postprocessing=statement_dict['post'],
    )




from typing import Tuple
from galaxy2janis.gx.configfiles.Configfile import Configfile

from ..TextRender import TextRender


class ConfigfileText(TextRender):
    def __init__(self, entity: Configfile):
        super().__init__()
        self.entity = entity

    @property
    def imports(self) -> list[Tuple[str, str]]:
        return []

    def render(self) -> str:
        return self.entity.contents
    




import os
import shutil
from typing import Any, Optional

from galaxy2janis.logs import logging
from galaxy2janis import settings
from galaxy2janis import fileio
from galaxy2janis import paths
from galaxy2janis import datatypes


def general_setup(args: dict[str, Any]) -> None:
    settings.general.set_command(args['command']) 

    parent_outdir = format_parent_outdir(args)
    tool_outdir = format_tool_subdir(args)

    paths.set_outdir(parent_outdir)
    paths.set_tool_subdir(tool_outdir)

    setup_output_folder(args)
    setup_data_folder()
    datatypes.populate()
    logging.configure_logging()


def setup_data_folder() -> None:
    # get package data folder path
    package_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    package_data_dir = f'{package_dir}/data'

    # create user data folder if not exists
    if not os.path.exists(paths.USER_DATA_DIR):
        os.mkdir(paths.USER_DATA_DIR)
    
    # create user wrapper download folder if not exists
    if not os.path.exists(f'{paths.USER_DATA_DIR}/{paths.DOWNLOADED_WRAPPERS_DIR}'):
        os.mkdir(f'{paths.USER_DATA_DIR}/{paths.DOWNLOADED_WRAPPERS_DIR}')

    # move files from package to user if they don't exist
    data_files = [
        paths.LOGGING_CONFIG,
        paths.GALAXY_CONFIG,
        paths.GALAXY_DATATYPES_YAML,
        paths.CONTAINER_CACHE,
        paths.WRAPPER_CACHE,
    ]
    for f in data_files:
        package_path = f'{package_data_dir}/{f}'
        user_path = f'{paths.USER_DATA_DIR}/{f}'
        if not os.path.exists(user_path):
            shutil.copyfile(package_path, user_path)

def format_tool_subdir(args: dict[str, Any]) -> Optional[str]:
    if args['command'] == 'tool':
        return None
    else:
        return 'tools'

def format_parent_outdir(args: dict[str, Any]) -> str:
    # user specified outdir
    if args['outdir']:
        return args['outdir']
    # auto formatted outdir
    else:
        parent_dir = auto_parent_dir(args)
        project_name = auto_project_name(args)
        return f'{parent_dir}/{project_name}'
    
def auto_parent_dir(args: dict[str, Any]) -> str:
    assert(args['outdir'] is None)
    match args['command']:
        case 'tool':
            return paths.DEFAULT_TOOL_OUTDIR
        case 'workflow':
            return paths.DEFAULT_WORKFLOW_OUTDIR
        case _:
            raise RuntimeError()

def auto_project_name(args: dict[str, Any]) -> str:
    match args['command']:
        case 'tool':
            return auto_tool_project_name(args)
        case 'workflow':
            return auto_workflow_project_name(args)
        case _:
            raise RuntimeError()

def auto_tool_project_name(args: dict[str, Any]) -> str:
    if args['infile']:
        return args['infile'].rsplit('/', 1)[-1].split('.', 1)[0]
    elif args['remote']:
        return args['remote'].split(',')[2]
    else:
        raise RuntimeError()

def auto_workflow_project_name(args: dict[str, Any]) -> str:
    return args['infile'].rsplit('/', 1)[-1].split('.', 1)[0]

def setup_output_folder(args: dict[str, Any]) -> None:
    # create outdir
    fileio.init_folder(paths.outdir())
    # create common subfolders
    for subfolder in paths.common_subfolders():
        fileio.init_folder(subfolder)
    # create workflow specific subfolders
    if args['command'] == 'workflow':
        for subfolder in paths.workflow_subfolders():
            fileio.init_folder(subfolder)








class ParamNotSupportedError(Exception):
    """Raise when the tool XML possesses a non-supported param"""

class TagNotSupportedError(Exception):
    """Raise when the tool XML possesses a non-supported XML tag"""

class AttributeNotSupportedError(Exception):
    """Raise when the tool XML possesses a non-supported XML tag"""

class DuplicateParamError(Exception):
    """Raise when two param names clash - another param with the same name was already known"""

class InputError(Exception):
    """Raise when user inputs are not complete / correct"""







from typing import Any, Optional
from .strategies import format_tag


class TagGroup:

    def __init__(self, section: str):
        self.section = section
        self.uuids_basetags: dict[str, str] = {}
        self.basetags_uuids: dict[str, list[str]] = {}
        self.active: bool = False

    def exists(self, uuid: str) -> bool:
        if uuid in self.uuids_basetags:
            return True
        return False
    
    def get(self, uuid: str) -> Optional[str]:
        if uuid in self.uuids_basetags:
            basetag = self.uuids_basetags[uuid]
            return self._generate_tag(basetag, uuid)
    
    def get_base_tag(self, uuid: str) -> str:
        return self.uuids_basetags[uuid]

    def register(self, starting_text: str, entity: Any) -> None:
        basetag = format_tag(starting_text, entity)
        uuid = entity.uuid

        self.uuids_basetags[uuid] = basetag
        if basetag not in self.basetags_uuids:
            self.basetags_uuids[basetag] = []
        self.basetags_uuids[basetag].append(uuid)

    def _generate_tag(self, basetag: str, query_uuid: str) -> Optional[str]:
        stored_uuids = self.basetags_uuids[basetag]
        if len(stored_uuids) <= 1:
            return basetag # only 1 object using this basetag
        for i, uuid in enumerate(stored_uuids):
            if uuid == query_uuid:
                return f'{basetag}{i+1}' # appends '1', '2' etc if basetag is shared by multiple objects
        return None





from typing import Any

from galaxy2janis.entities.tool import Tool
from galaxy2janis.entities.workflow import StepMetadata
from galaxy2janis.entities.workflow import Workflow

from galaxy2janis.tool_mode import tool_mode
from galaxy2janis.gx.gxworkflow.parsing.tool_state import load_tool_state
from galaxy2janis.gx.interaction import get_builtin_tool_path

from galaxy2janis import mapping
from galaxy2janis import settings

def ingest_workflow_tools(janis: Workflow, galaxy: dict[str, Any]) -> None:
    for g_step in galaxy['steps'].values():
        if g_step['type'] == 'tool':
            j_step = mapping.step(g_step['id'], janis, galaxy)
            tool = parse_step_tool(j_step.metadata)
            j_step.set_tool(tool)
            g_step['tool_state'] = load_tool_state(g_step)

def parse_step_tool(metadata: StepMetadata) -> Tool:
    args = create_tool_settings_for_step(metadata)
    settings.tool.set(args)
    return tool_mode()

def create_tool_settings_for_step(metadata: StepMetadata) -> dict[str, Any]:
    tool_id = metadata.wrapper.tool_id
    if metadata.wrapper.inbuilt:
        xml_path = get_builtin_tool_path(tool_id)
        assert(xml_path)
        return {
            'infile': xml_path,
            'remote': None,
            'outdir': None
            #'outdir': f'{paths.wrapper(tool_id, tool_id)}'
        }
    else:
        revision = metadata.wrapper.revision
        owner = metadata.wrapper.owner
        repo = metadata.wrapper.repo
        return {
            'infile': None,
            'remote': f'{owner},{repo},{tool_id},{revision}',
            'outdir': None
            #'outdir': f'{paths.wrapper(tool_id, revision)}'
        }





from dataclasses import dataclass
from typing import Optional
from uuid import uuid4

from galaxy2janis import tags
from galaxy2janis.gx.configfiles.Configfile import Configfile
from galaxy2janis.gx.gxtool import ToolXMLMetadata
from galaxy2janis.gx.gxtool.param import ParamRegister
from galaxy2janis.gx.gxtool.param import Param
from galaxy2janis.gx.command.components import CommandComponent
from galaxy2janis.gx.command.components import InputComponent
from galaxy2janis.gx.command.components import OutputComponent


# shouldn't really be a dataclass, just annoying to write __init__ method
@dataclass
class Tool:
    """
    a Tool() is the final representation of the software tool
    a galaxy XML wrapper is running. Includes metadata, inputs, outputs, a container to execute the tool, base command etc. 
    """
    metadata: ToolXMLMetadata
    gxparam_register: ParamRegister
    configfiles: list[Configfile]
    container: Optional[str]
    base_command: list[str]

    def __post_init__(self):
        self.inputs: list[InputComponent] = []
        self.outputs: list[OutputComponent] = []
        self.uuid: str = str(uuid4())
        tags.new_group('tool', self.uuid)
        tags.register(self)

    @property
    def name(self) -> str:
        return self.metadata.id
    
    @property
    def tag(self) -> str:
        return tags.get(self.uuid)

    def add_input(self, inp: InputComponent) -> None:
        tags.switch_group(self.uuid)
        tags.register(inp)
        self.inputs.append(inp)
    
    def add_output(self, out: OutputComponent) -> None:
        tags.switch_group(self.uuid)
        tags.register(out)
        self.outputs.append(out)

    def get_gxparam(self, query: str) -> Optional[Param]:
        param = self.gxparam_register.get(query, strategy='lca')
        if not param:
            pass
            #raise RuntimeError(f'no gxparam named {query}')
        return param
   
    def get_input(self, query_uuid: str) -> Optional[CommandComponent]:
        for inp in self.inputs:
            if query_uuid == inp.uuid:
                return inp
        raise RuntimeError(f'could not find {query_uuid} in tool inputs')
    
    def get_input_via_param_name(self, query_uuid: str) -> CommandComponent:
        for inp in self.inputs:
            if query_uuid == inp.uuid:
                return inp
        raise RuntimeError(f'could not find {query_uuid} in tool inputs')

    def get_preprocessing(self) -> Optional[str]:
        raise NotImplementedError

    def get_postprocessing(self) -> Optional[str]:
        raise NotImplementedError



    





from typing import Any

from galaxy2janis.entities.workflow import Workflow
from galaxy2janis.entities.workflow import WorkflowStep

from .tool_step.metadata import parse_step_metadata


def ingest_workflow_steps(janis: Workflow, galaxy: dict[str, Any]) -> None:
    for step in galaxy['steps'].values():
        if step['type'] == 'tool':
            workflow_step = parse_tool_step(step)
            step['janis_uuid'] = workflow_step.uuid # crucial!
            janis.add_step(workflow_step)

def parse_tool_step(step: dict[str, Any]) -> WorkflowStep:
    metadata = parse_step_metadata(step)
    return WorkflowStep(metadata=metadata)




import json
import os
from typing import Optional
from galaxy2janis.gx.wrappers import Wrapper
from galaxy2janis.paths import USER_DATA_DIR, WRAPPER_CACHE

"""
WrapperCache flat file structure:
{
    tool_id: [
            {
                'owner': 'fastqc',
                'repo': 'fastqc',
                'revision': 'fastqc',
                'tool_id': 'fastqc',
                'tool_version': 'fastqc',
                'tool_build': 'fastqc',
                'date_created': [revision date],
                'url': [toolshed download link],
            },
            ...
        ]
    }
}
"""



class WrapperCache:

    def get(
        self, 
        tool_id: Optional[str]=None,
        tool_build: Optional[str]=None,
        owner: Optional[str]=None,
        repo: Optional[str]=None,
        revision: Optional[str]=None
    ) -> list[Wrapper]:
        """loads cache then returns each Wrapper satisfying the query"""
        # load cache and get relevant wrapper in dict form
        cache = self._load()

        wrappers: list[dict[str, str]] = []
        for wrapper_list in cache.values():
            wrappers += wrapper_list
        
        # this can be more elegant ofc
        if tool_id:
            wrappers = [x for x in wrappers if x['tool_id'] == tool_id]
        if tool_build:
            wrappers = [x for x in wrappers if x['tool_build'] == tool_build]
        if owner:
            wrappers = [x for x in wrappers if x['owner'] == owner]
        if repo:
            wrappers = [x for x in wrappers if x['repo'] == repo]
        if revision:
            wrappers = [x for x in wrappers if x['revision'] == revision]
        
        # cast to Wrapper instances
        return [Wrapper(w) for w in wrappers]
    
    def add(self, wrapper: Wrapper) -> None:
        """adds the Wrapper to our cache and saves to file"""
        # load cache 
        cache = self._load() 

        # alter cache if needed
        self._ensure_key(wrapper.tool_id, cache)
        if not self.exists(wrapper, cache):
            cache[wrapper.tool_id].append(wrapper.to_dict())
        
        # write
        self._save(cache) 

    def exists(self, wrapper: Wrapper, cache: dict[str, list[dict[str, str]]]) -> bool:
        """checks if the wrapper is already in cache"""
        if wrapper.tool_id in cache:
            for entry in cache[wrapper.tool_id]:
                if entry['revision'] == wrapper.revision:
                    return True
        return False

    def _load(self) -> dict[str, list[dict[str, str]]]:
        """loads cache from flat file"""
        path = f'{USER_DATA_DIR}/{WRAPPER_CACHE}'
        if not os.path.exists(path):
            with open(path, 'w') as fp:
                fp.write('{}')
        with open(path, 'r') as fp:
            return json.load(fp)
    
    def _save(self, cache: dict[str, list[dict[str, str]]]) -> None:
        """saves our cache to file"""
        path = f'{USER_DATA_DIR}/{WRAPPER_CACHE}'
        with open(path, 'w') as fp:
            json.dump(cache, fp)

    def _ensure_key(self, key: str, cache: dict[str, list[dict[str, str]]]) -> None:    
        """ensures { key: [] } in cache"""
        if key not in cache:
            cache[key] = []
            
    




"""

module info

main calls and operations here using ToolshedScraper etc

"""

import sys
sys.path.append('./src')

from galaxy2janis.gx.wrappers.scraping.repositories import scrape_repos
from galaxy2janis.gx.wrappers.scraping.revisions import scrape_revisions
from galaxy2janis.gx.wrappers.scraping.wrappers import scrape_wrappers


def main(argv: list[str]):
    mode = argv[0]
    match mode:
        case 'repos':
            scrape_repos()
        case 'revisions':
            scrape_revisions()
        case 'wrappers':
            scrape_wrappers()
        case _:
            print(f'invalid mode {mode}')
            sys.exit()


if __name__ == '__main__':
    main(sys.argv[1:])







from typing import Any
from galaxy2janis.gx.gxtool.requirements import Requirement

from ..Container import Container
from ..fetching.Fetcher import Fetcher
from . import utils


class QuayIOFetcher(Fetcher):

    def fetch(self, requirement: Requirement) -> list[Container]:
        interactor = QuayInteractor()
        tool_data = interactor.request_tool_data(requirement.name)
        if tool_data:
            return self.response_to_containers(tool_data, requirement)
        else:
            return []

    def response_to_containers(self, repo: dict[str, Any], requirement: Requirement) -> list[Container]:
        out: list[Container] = []
        for details in repo['tags'].values():
            info: dict[str, str] = {
                'image_type': 'docker',
                'repo': repo['name'],
                'tag': details['name'],
                'url': self.format_url(repo, details),
                '_timestamp': details['last_modified']
            }
            out.append(Container(info))
        return out
    
    def format_url(self, repo: dict[str, Any], tag: dict[str, str]) -> str:
        return f"quay.io/biocontainers/{repo['name']}:{tag['name']}"


class QuayInteractor:

    def request_tool_data(self, name: str) -> dict[str, Any]:
        endpoint = f'https://quay.io/api/v1/repository/biocontainers/{name}'
        return utils.make_api_request(endpoint)


    







from janis_core.tool.test_classes import TTestCase


class TestRegister:
    def __init__(self, gxtests: list[TTestCase]):
        self.tests: dict[str, TTestCase] = {t.name: t for t in gxtests}

    def list(self) -> list[TTestCase]:
        return list(self.tests.values())









from ..gxtool.XMLToolDefinition import XMLToolDefinition
from ..gxtool.text import load_command

from .cmdstr.CommandString import CommandString
from .cmdstr.generate import gen_command_string
from .annotation import ArgumentCommandAnnotator
from .annotation import CmdstrCommandAnnotator
from .Command import Command


class CommandFactory:
    def __init__(self, xmltool: XMLToolDefinition):
        self.xmltool = xmltool
        self.xmlcmdstr = self.gen_cmdstr_from_xml()
        self.command = Command(self.xmlcmdstr)

    def create(self) -> Command:
        self.update_command_via_arguments()
        self.update_command_via_cmdstrs()
        return self.command

    def update_command_via_arguments(self) -> None:
        """uses galaxy params with an 'argument' attribute to update command"""
        annotator = ArgumentCommandAnnotator(self.command, self.xmltool)
        annotator.annotate()
    
    def update_command_via_cmdstrs(self) -> None:
        """
        uses valid command line strings from tests, and the tool XML <command> section
        to further identify the structure and options of the underling software tool
        """
        # create command strings (from evaluated tests, simplified xml <command>)
        cmdstrs = self.gen_cmdstrs()
        annotator = CmdstrCommandAnnotator(self.command, self.xmltool, cmdstrs)
        annotator.annotate()

    def gen_cmdstrs(self) -> list[CommandString]:
        # note ordering: xml then test
        return [self.xmlcmdstr]

    def gen_cmdstr_from_xml(self) -> CommandString:
        text = load_command()
        return gen_command_string(source='xml', the_string=text, xmltool=self.xmltool)








from abc import ABC, abstractmethod
from typing import Tuple
from galaxy2janis.entities.workflow.step.inputs import ConnectionInputValue, InputValue, WorkflowInputInputValue

from galaxy2janis.gx.command.components import Flag
from galaxy2janis.gx.command.components import Option
from galaxy2janis.gx.command.components import Positional



def order_positionals(positionals: list[Positional]) -> list[Positional]:
    positionals.sort(key=lambda x: x.cmd_pos)
    return positionals

def order_flags(flags: list[Flag]) -> list[Flag]:
    flags.sort(key=lambda x: x.tag)
    return flags

def order_options(options: list[Option]) -> list[Option]:
    options.sort(key=lambda x: x.tag)
    return options

def order_imports(imports: list[Tuple[str, str]]) -> list[Tuple[str, str]]:
    imports = _order_imports_alphabetical(imports)
    imports = _order_imports_length(imports)
    return imports

def _order_imports_length(imports: list[Tuple[str, str]]) -> list[Tuple[str, str]]:
    imports.sort(key=lambda x: len(x[0] + x[1]), reverse=True)
    return imports

def _order_imports_alphabetical(imports: list[Tuple[str, str]]) -> list[Tuple[str, str]]:
    imports.sort(key=lambda x: f'from {x[0]} import {x[1]}')
    return imports




### INPUT VALUES ###

class InputOrderingStrategy(ABC):
    @abstractmethod
    def order(self, invalues: list[InputValue]) -> list[InputValue]:
        """orders input values and returns ordered list"""
        ...

class AlphabeticalStrategy(InputOrderingStrategy):
    def order(self, invalues: list[InputValue]) -> list[InputValue]:
        invalues.sort(key=lambda x: x.input_tag)
        return invalues

class NotNullPriority(InputOrderingStrategy):
    def order(self, invalues: list[InputValue]) -> list[InputValue]:
        invalues.sort(key=lambda x: x.input_value != 'None', reverse=True)
        return invalues

class PositionalsOptsPositionals(InputOrderingStrategy):
    def order(self, invalues: list[InputValue]) -> list[InputValue]:
        opts_pos = self.get_opts_position(invalues)
        top: list[InputValue] = []
        middle: list[InputValue] = []
        bottom: list[InputValue] = []
        for inval in invalues:
            if inval.component and inval.comptype == 'positional':
                if inval.component.cmd_pos < opts_pos:
                    top.append(inval)
                else:
                    bottom.append(inval)
            else:
                middle.append(inval)
        return top + middle + bottom

    def get_opts_position(self, invalues: list[InputValue]) -> int:
        for inval in invalues:
            if inval.component and inval.comptype != 'positional':
                return inval.component.cmd_pos
        return 1

class RuntimeInputPriorityStrategy(InputOrderingStrategy):
    def order(self, invalues: list[InputValue]) -> list[InputValue]:
        invalues.sort(key=lambda x: isinstance(x, WorkflowInputInputValue) and x.is_runtime, reverse=True)
        return invalues

class WorkflowInputPriorityStrategy(InputOrderingStrategy):
    def order(self, invalues: list[InputValue]) -> list[InputValue]:
        invalues.sort(key=lambda x: isinstance(x, WorkflowInputInputValue), reverse=True)
        return invalues

class ConnectionPriorityStrategy(InputOrderingStrategy):
    def order(self, invalues: list[InputValue]) -> list[InputValue]:
        invalues.sort(key=lambda x: isinstance(x, ConnectionInputValue), reverse=True)
        return invalues

class UnlinkedPriorityStrategy(InputOrderingStrategy):
    def order(self, invalues: list[InputValue]) -> list[InputValue]:
        invalues.sort(key=lambda x: x.component is not None)
        return invalues


# changing the order of the objects below changes the 
# ordering priority, as the last ordering method has the highest impact etc

"""
- step tool value order:
    - connection 
    - wflow input 
    - runtime
    - positionals (before opts) 
    - things with values
        - flags 
        - options 
    - things without values
        - flags 
        - options 
    - positionals (after opts)
    - each category above should appear alphabetically
"""

STRATEGIES = [
    AlphabeticalStrategy(),
    NotNullPriority(),
    PositionalsOptsPositionals(),
    RuntimeInputPriorityStrategy(),
    ConnectionPriorityStrategy(),
    WorkflowInputPriorityStrategy(),
    UnlinkedPriorityStrategy()
]

def order_step_inputs(invalues: list[InputValue]):
    for strategy in STRATEGIES:
        invalues = strategy.order(invalues)
    return invalues




from galaxy2janis.logs import logging
from galaxy2janis.gx.command import Command
from galaxy2janis.gx.command.components import CommandComponent
from galaxy2janis.gx.command.components import Flag
from galaxy2janis.gx.command.components import RedirectOutput
from galaxy2janis.gx.command.components import factory
from galaxy2janis.gx.gxtool import XMLToolDefinition
from galaxy2janis.gx.gxtool.param import Param


def extract_outputs(xmltool: XMLToolDefinition, command: Command) -> list[CommandComponent]:
    extractor = OutputExtractor(xmltool, command)
    return extractor.extract()


class OutputExtractor:
    def __init__(self, xmltool: XMLToolDefinition, command: Command):
        self.xmltool = xmltool
        self.command = command

    def extract(self) -> list[CommandComponent]:
        outputs: list[CommandComponent] = []
        outputs += self.get_redirect_outputs()
        outputs += self.get_input_outputs()
        outputs += self.get_wildcard_outputs()
        outputs += self.get_uncertain_outputs(outputs)
        return outputs
        
    def get_redirect_outputs(self) -> list[CommandComponent]:
        # redirect outputs (stdout) already identified when creating Command()
        # need to ensure they're linked to a gxparam
        # if not, try to link to dataset collector, else just ignore as the 
        # redirect seems to be dropped.
        out: list[CommandComponent] = []
        redirects: list[RedirectOutput] = self.command.list_outputs()
        for r in redirects:
            self.attempt_redirect_gxparam_link(r)
            if r.gxparam is not None:
                out.append(r)
        return out

    def attempt_redirect_gxparam_link(self, r: RedirectOutput) -> None:
        if not r.gxparam:
            for query_param in self.xmltool.outputs.list():
                if query_param.discover_pattern is not None:
                    if query_param.discover_pattern == r.values.most_common_value:
                        r.gxparam = query_param

    def get_input_outputs(self) -> list[CommandComponent]:
        # can be identified by looking at the input components which
        # have attached gxparams which are outputs. 
        # example case: toolname input.fastq -o $outfile
        # the -o option would be picked up as a command component, and the
        # gxparam referred to by $outfile is stored on the command component
        out: list[CommandComponent] = []
        for component in self.command.list_inputs(include_base_cmd=False):
            if self.should_create_input_output(component):
                output = factory.input_output(component)
                out.append(output)
        return out
    
    def get_wildcard_outputs(self) -> list[CommandComponent]:
        # verified vs unverified:
        # only if no post-processing! otherwise, wildcard outputs may 
        # have come from post-processing.
        
        # outputs which were not identified in the command
        # usually just because they have a file collection strategy
        # like from_work_dir or a <discover_datatsets> tag as a child
        out: list[CommandComponent] = []
        for gxparam in self.xmltool.outputs.list():
            if self.should_create_wildcard_output(gxparam):
                output = factory.wildcard_output(gxparam)
                out.append(output)
        return out

    def get_uncertain_outputs(self, existing_outputs: list[CommandComponent]) -> list[CommandComponent]:
        out: list[CommandComponent] = []
        for gxparam in self.xmltool.outputs.list():
            if self.should_create_uncertain_output(gxparam, existing_outputs):
                logging.uncertain_output()
                output = factory.uncertain_output(gxparam)
                out.append(output)
        return out

    # CHECKS
    def should_create_input_output(self, component: CommandComponent) -> bool:
        if not isinstance(component, Flag):
            if component.gxparam and self.xmltool.outputs.get(component.gxparam.name):
                return True
        return False

    def should_create_wildcard_output(self, gxparam: Param) -> bool:
        """test to see if this *galaxy output param* should spawn WildcardOutput"""
        if not self.command.gxparam_is_attached(gxparam):
            if hasattr(gxparam, 'from_work_dir') and gxparam.from_work_dir is not None: # type: ignore
                return True
            elif hasattr(gxparam, 'discover_pattern') and gxparam.discover_pattern is not None: # type: ignore
                return True
        return False

    def should_create_uncertain_output(self, gxparam: Param, existing_outputs: list[CommandComponent]) -> bool:
        """
        test to see if this *galaxy output param* should spawn uncertain WildcardOutput
        all galaxy outputs which are not yet accounted for must become uncertain WildcardOutputs
        """
        has_output_component = False
        for output in existing_outputs:
            if output.gxparam and output.gxparam.name == gxparam.name:
                has_output_component = True
                break
        if not has_output_component:
            return True
        return False

    def verify_outputs(self, outputs: list[CommandComponent]) -> None:
        # just checks we have the same number of outputs identified as CommandComponents
        # as there are in the xmltool's listed output params
        assert(len(self.xmltool.outputs.list()) == len(outputs))




import unittest
import xml.etree.ElementTree as et

from gx.gxtool.text.simplification.aliases import resolve_aliases



def get_cmd(path: str) -> str:
    tree = et.parse(path)
    root = tree.getroot()
    assert(root.text)
    return root.text


class TestAliases(unittest.TestCase):

    def test_resolve_fastqc(self):
        raw_path = 'test/data/command/manipulation/aliases/fastqc/fastqc_command.xml'
        ref_path = 'test/data/command/manipulation/aliases/fastqc/fastqc_command_resolved.xml'
        raw_cmd = get_cmd(raw_path)
        ref_cmd = get_cmd(ref_path)
        res_cmd = resolve_aliases(raw_cmd)
        self.assertEquals(ref_cmd, res_cmd)
    
    def test_resolve_unicycler(self):
        raw_path = 'test/data/command/manipulation/aliases/unicycler/unicycler_command.xml'
        ref_path = 'test/data/command/manipulation/aliases/unicycler/unicycler_command_resolved.xml'
        raw_cmd = get_cmd(raw_path)
        ref_cmd = get_cmd(ref_path)
        res_cmd = resolve_aliases(raw_cmd)
        self.assertEquals(ref_cmd, res_cmd)
        






from typing import Tuple

from galaxy2janis.entities.workflow import WorkflowInput

from ..TextRender import TextRender
from .. import ordering
from .. import formatting


class WorkflowInputText(TextRender):
    def __init__(self, entity: WorkflowInput):
        super().__init__()
        self.entity = entity

    @property
    def imports(self) -> list[Tuple[str, str]]:
        jtype = self.entity.datatype
        imports: list[Tuple[str, str]] = []
        imports.append((jtype.import_path, jtype.classname))

        # TODO opportunity for decorator
        if self.entity.array:
            imports.append(('janis_core', 'Array'))

        # TODO opportunity for decorator
        imports = list(set(imports))
        return ordering.order_imports(imports)

    def render(self) -> str:
        datatype_str = formatting.format_typestr(self.entity)
        out_str: str = ''
        out_str += f'w.input("{self.entity.tag}", {datatype_str})'
        #out_str += f'\t"{tag}",\n'
        #out_str += f'\t{datatype}'
        #out_str += f',\n\tdefault={default}' if default else ''  # TODO HERE
        #out_str += f',\n\tvalue={value}' if value else ''
        #out_str += f',\n\tdoc="{doc}"' if doc else ''
        #out_str += '\n)\n'
        return out_str




from __future__ import annotations
from typing import TYPE_CHECKING

from galaxy2janis.fileio.text.workflow.InputsText import InputsText

if TYPE_CHECKING:
    from galaxy2janis.entities.tool import Tool
    from galaxy2janis.entities.workflow import Workflow
    from galaxy2janis.entities.workflow import WorkflowStep

import shutil
from galaxy2janis import paths

from galaxy2janis.utils import galaxy as galaxy_utils
from galaxy2janis.gx.wrappers import fetch_wrapper

from .text.workflow.WorkflowText import WorkflowText
from .text.tool.ConfigfileText import ConfigfileText
from .text.tool.UnstranslatedText import UntranslatedText
from .text.tool.ToolText import ToolText

from .initialisation import init_folder




def write_tool(tool: Tool, path: str) -> None:
    text = ToolText(tool)
    page = text.render()
    with open(path, 'w') as fp:
        fp.write(page)

def write_workflow(janis: Workflow) -> None:
    write_tools(janis)
    write_untranslated(janis)
    write_scripts(janis)
    write_wrappers(janis)
    write_main_workflow(janis)
    write_inputs(janis)
    #write_sub_workflows(janis)
    #write_config(janis)

def write_tools(janis: Workflow) -> None:
    for step in janis.steps:
        tool_id = step.metadata.wrapper.tool_id
        write_tool(step.tool, paths.tool(tool_id))

def write_untranslated(janis: Workflow) -> None:
    for step in janis.steps:
        if step.preprocessing or step.postprocessing:
            tool_id = step.metadata.wrapper.tool_id
            path = paths.untranslated(tool_id)
            text = UntranslatedText(step)
            page = text.render()
            with open(path, 'w') as fp:
                fp.write(page)

def write_scripts(janis: Workflow) -> None:
    for step in janis.steps:
        if step.tool.configfiles:
            tool_id = step.metadata.wrapper.tool_id
            for configfile in step.tool.configfiles:
                path = paths.configfile(tool_id, configfile.name)
                text = ConfigfileText(configfile)
                page = text.render()
                with open(path, 'w') as fp:
                    fp.write(page)

def write_wrappers(janis: Workflow) -> None:
    for step in janis.steps:
        src_files = get_wrapper_files_src(step)
        dest = get_wrapper_files_dest(step)
        init_folder(dest)
        for src in src_files:
            shutil.copy2(src, dest)

def get_wrapper_files_src(step: WorkflowStep) -> list[str]:
    wrapper = step.metadata.wrapper
    wrapper_path = fetch_wrapper(
        owner= wrapper.owner,
        repo= wrapper.repo,
        revision= wrapper.revision,
        tool_id= wrapper.tool_id
    )
    wrapper_dir = wrapper_path.rsplit('/', 1)[0]
    macro_xmls = galaxy_utils.get_macros(wrapper_dir)
    return [wrapper_path] + macro_xmls

def get_wrapper_files_dest(step: WorkflowStep) -> str:
    tool_id = step.metadata.wrapper.tool_id
    revision = step.metadata.wrapper.revision
    return paths.wrapper(tool_id, revision)

def write_main_workflow(janis: Workflow) -> None:
    path = paths.workflow()
    text = WorkflowText(janis)
    page = text.render()
    with open(path, 'w') as fp:
        fp.write(page)

def write_inputs(janis: Workflow) -> None:
    path = paths.inputs(file_format='yaml')
    text = InputsText(janis, file_format='yaml')
    page = text.render()
    with open(path, 'w') as fp:
        fp.write(page)

def write_sub_workflows(janis: Workflow) -> None:
    raise NotImplementedError()

def write_config(janis: Workflow) -> None:
    raise NotImplementedError()








from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Tuple

from galaxy2janis.messages import Message


class CurationState(Enum):
    RAW     = auto()
    EDITED  = auto()
    CURATED = auto()

class Confidence(Enum):
    LOW     = auto()
    MEDIUM  = auto()
    HIGH    = auto()


class TextRender(ABC):
    """
    A TextRender is a text representation of an entity (tool input, workflow step etc).
    Similar to __str__ or __repr__ dunder methods but not coupled to the class. 
    Includes other information like imports and messages related to that entity. 
    Used to create the output definitions users will receive. 
    \n
    Accounts for all the different ways the entity could be rendered. 
    Example 1: whether or not imports will be rendered here
    Example 2: how messages are 
    displayed to the user etc. 
    """

    def __init__(self):
        self.messages: list[Message] = []
        self.curation_state: CurationState = CurationState.RAW
        self.confidence: Confidence = Confidence.LOW

    @property
    @abstractmethod
    def imports(self) -> list[Tuple[str, str]]:
        """
        collects all imports this entity will need.
        each import should be returned as [path(module), object]. 
        """
        ...

    @abstractmethod
    def render(self) -> str:
        """renders an entity to text"""
        ...
    









from __future__ import annotations
from dataclasses import dataclass
from typing import Optional

from .DynamicCommandStatement import DynamicCommandStatement
from galaxy2janis.utils.general import global_align


# functions to identify best statement
def undefined_future_function(metric_records: list[StatementMetricRecord]) -> Optional[int]:
    # TODO statment with most flags? statement with redirect? statement firstword is not in list of known linux commands (except in cases like where the tool is actually 'awk')?
    raise NotImplementedError()

def perfect_requirement_match(metric_records: list[StatementMetricRecord]) -> Optional[int]:
    metric_records.sort(key=lambda x: x.reqsim, reverse=True)
    if metric_records[0].reqsim == 1 and metric_records[1].reqsim < 1:
        return metric_records[0].index
    return None

def top_all_metrics(metric_records: list[StatementMetricRecord]) -> Optional[int]:
    # same statement at the top of both metrics
    best_reqsim = sorted(metric_records, key=lambda x: x.reqsim, reverse=True)[0]
    best_gxrefs = sorted(metric_records, key=lambda x: x.gxrefs, reverse=True)[0]
    if best_reqsim.gxrefs == best_gxrefs.gxrefs:
        return best_reqsim.index
    return None

def requirement_priority(metric_records: list[StatementMetricRecord]) -> Optional[int]:
    # if 1+ statments firstword is very similar to the tool id (> 80% similar)
    # return the statement in this group with most gxrefs
    candidates = [x for x in metric_records if x.reqsim > 0.8]
    if candidates:
        return sorted(candidates, key=lambda x: x.gxrefs, reverse=True)[0].index
    return None

def gxref_priority(metric_records: list[StatementMetricRecord]) -> Optional[int]:
    # one statment has at least 3 galaxy references, and this is 2x more than others
    metric_records.sort(key=lambda x: x.gxrefs, reverse=True)
    if metric_records[0].gxrefs >= 3 and metric_records[0].gxrefs >= 2 * metric_records[1].gxrefs:
        return metric_records[0].index
    return None



@dataclass
class StatementMetricRecord:
    index: int
    statement: DynamicCommandStatement
    gxrefs: int = 0 # the number of references to galaxy params in the statement
    reqsim: float = 0 # the similarity between the main tool requirement, and the first word of the statement

class MainStatementInferrer:
    def __init__(self, statements: list[DynamicCommandStatement], source: str, requirement: str):
        self.statements = statements
        self.source = source
        self.requirement = requirement
        self.metric_records: list[StatementMetricRecord] = []

    def infer(self) -> int:
        if len(self.statements) <= 1:
            return 0  # fallback: first statement 

        self.init_metric_record()
        self.set_gxref_counts()
        self.set_mainreq_similarities()
        if self.source == 'xml':
            best = self.choose_best_statement_xml()
        if self.source == 'test':
            best = self.choose_best_statement_test()
        return best
        
    def init_metric_record(self) -> None:
        for i, statement in enumerate(self.statements):
            new_record = StatementMetricRecord(i, statement)
            self.metric_records.append(new_record)
        
    def set_gxref_counts(self) -> None:
        for record in self.metric_records:
            record.gxrefs = record.statement.get_galaxy_reference_count()
        
    def set_mainreq_similarities(self) -> None:
        banned_command_starters = ['cp', 'ln', 'mv']
        max_possible_score = global_align(self.requirement, self.requirement)
        for record in self.metric_records:
            if len(record.statement.realised_tokens) == 0:
                record.reqsim = 0
            elif record.statement.get_first_word() in banned_command_starters:
                record.reqsim = -99999.9 # HACK
            else:
                raw_similarity = global_align(record.statement.get_first_word(), self.requirement)
                record.reqsim = raw_similarity / max_possible_score
    
    def choose_best_statement_xml(self) -> int:
        identifiers = [
            top_all_metrics,
            perfect_requirement_match,
            requirement_priority,
            gxref_priority
        ]
        for identifier_func in identifiers:
            selection = identifier_func(self.metric_records)
            if selection is not None: 
                return selection
        return len(self.statements) - 1 # fallback: the last statement

    def choose_best_statement_test(self) -> int:
        identifiers = [
            perfect_requirement_match,
            top_all_metrics,
            requirement_priority,
            #gxref_priority
        ]
        for identifier_func in identifiers:
            selection = identifier_func(self.metric_records)
            if selection is not None: 
                return selection
        return len(self.statements) - 1 # fallback: the last statement

      






from typing import Protocol
from galaxy2janis.datatypes import JanisDatatype


class TaggedEntity(Protocol):
    """
    A workflow entity which has a tag.
    Includes Tools, WorkflowSteps, Workflow, WorkflowInput, ToolInputs, ToolOutputs.
    """
    uuid: str
    
    @property
    def name(self) -> str:
        """name of this entity. Derived from held information"""
        ...
    
    @property
    def tag(self) -> str:
        """provides the current tag for this entity"""
        ...



class TypedEntity(Protocol):
    """
    A workflow entity which has a datatype.
    Includes WorkflowInput, WorkflowOutput, ToolInput, ToolOutput
    """

    @property
    def datatype(self) -> JanisDatatype:
        """
        name of this entity. 
        should be derived from information gathered about the entity
        """
        ...




from typing import Any, Optional
from .Param import Param


class OutputParam(Param):
    def __init__(self, name: str):
        self.name: str = name
        self.label: str = ''
        self.formats: list[str] = []
        self.discover_pattern: Optional[str] = None

    @property
    def default(self) -> Any:
        return None

    @property
    def docstring(self) -> str:
        if self.label:
            return str(self.label)
        return ''
    
    @property
    def optional(self) -> bool:
        return False
    
    @property
    def array(self) -> bool:
        return False


class DataOutputParam(OutputParam):
    def __init__(self, name: str, from_work_dir: Optional[str]=None, discover_pattern: Optional[str]=None):
        super().__init__(name)
        self.from_work_dir = from_work_dir
        self.discover_pattern = discover_pattern

    @property
    def array(self) -> bool:
        if self.discover_pattern and '*' in self.discover_pattern:
            return True
        return False


class CollectionOutputParam(OutputParam):
    def __init__(self, name: str, discover_pattern: Optional[str]=None):
        super().__init__(name)
        self.discover_pattern = discover_pattern
        self.collection_type: str = 'list'

    @property
    def array(self) -> bool:
        return True






from abc import ABC, abstractmethod
from typing import Optional
from galaxy.tool_util.parser.output_objects import ToolOutput as GxOutput

from ..param.ParamRegister import ParamRegister
from ..param.OutputParam import (
    OutputParam,
    DataOutputParam,
    CollectionOutputParam,
)

from galaxy2janis import expressions
from galaxy2janis.expressions.patterns import WILDCARD_GROUP


class Factory(ABC):
    @abstractmethod
    def create(self, gxout: GxOutput, inputs: ParamRegister)  -> OutputParam:
        """parses galaxy output to return an OutputParam"""
        ...

class DataOutputFactory(Factory):
    def create(self, gxout: GxOutput, inputs: ParamRegister)  -> OutputParam:
        param = DataOutputParam(gxout.name)
        param.label = str(gxout.label).rsplit('}', 1)[-1].strip(': ')
        param.formats = fetch_format(gxout, inputs)
        param.from_work_dir = get_from_workdir_pattern(gxout)
        param.discover_pattern = get_discover_pattern(gxout)
        return param

class CollectionOutputFactory(Factory):
    def create(self, gxout: GxOutput, inputs: ParamRegister)  -> OutputParam:
        param = CollectionOutputParam(gxout.name)
        param.label = str(gxout.label).rsplit('}', 1)[-1].strip(': ')
        if gxout.structure.collection_type != '':
            param.collection_type = str(gxout.structure.collection_type) 
        param.formats = fetch_format(gxout, inputs)
        param.discover_pattern = get_discover_pattern(gxout)
        return param


def get_from_workdir_pattern(gxout: GxOutput) -> Optional[str]:
    if has_from_workdir(gxout):
        return gxout.from_work_dir
    return None

def get_discover_pattern(gxout: GxOutput) -> Optional[str]:
    if has_dataset_collector(gxout):
        collector = gxout.dataset_collector_descriptions[0]
        pattern = f'{collector.directory}/{collector.pattern}' # type: ignore
        return remove_pattern_capture_groups(pattern)
    return None

def remove_pattern_capture_groups(pattern: str) -> str:
    matches = expressions.get_matches(pattern, WILDCARD_GROUP)
    for m in matches:
        pattern = pattern.replace(m.group(0), m.group(2))
    return pattern



factory_map = {
    'data': DataOutputFactory,
    'float': DataOutputFactory,
    'collection': CollectionOutputFactory
}

def parse_output_param(gxout: GxOutput, inputs: ParamRegister) -> list[OutputParam]:
    galaxy_params: list[GxOutput] = []
    internal_params: list[OutputParam] = []

    # split collection of defined outputs to list 
    if is_defined_collection(gxout):
        for g_param in gxout.outputs.values(): # type: ignore
            galaxy_params.append(g_param)
    # parse the output as usual
    else:
        galaxy_params.append(gxout)
    
    for g_param in galaxy_params:
        f_class = factory_map[g_param.output_type]
        factory = f_class()
        i_param = factory.create(g_param, inputs)
        internal_params.append(i_param)
    return internal_params

def is_defined_collection(gxout: GxOutput) -> bool:
    if hasattr(gxout, 'outputs') and len(gxout.outputs) > 0:  # type: ignore
        return True
    return False


def fetch_format(gxout: GxOutput, inputs: ParamRegister) -> list[str]:
    strategy: FetchStrategy = select_fetcher(gxout)
    return strategy.fetch(gxout, inputs)

def has_format(gxout: GxOutput) -> bool:
    if hasattr(gxout, 'format'):
        if gxout.format and gxout.format != 'data':
            return True
    return False

def has_default_format(gxout: GxOutput) -> bool:
    if hasattr(gxout, 'default_format'): 
        if gxout.default_format and gxout.default_format != 'data':
            return True
    return False

def has_format_source(gxout: GxOutput) -> bool:
    if hasattr(gxout, 'format_source'):
        if gxout.format_source:
            return True
    return False

def has_from_workdir(gxout: GxOutput) -> bool:
    if gxout.output_type == 'data': # type: ignore
        if hasattr(gxout, 'from_work_dir') and gxout.from_work_dir: # type: ignore
            return True
    return False

def has_dataset_collector(gxout: GxOutput) -> bool:
    if hasattr(gxout, 'dynamic_structure') and gxout.dynamic_structure: # type: ignore
        if hasattr(gxout, 'dataset_collector_descriptions'):
            return True
    return False

# helper classes 
class FetchStrategy(ABC):
    @abstractmethod
    def fetch(self, gxout: GxOutput, inputs: ParamRegister) -> list[str]:
        """gets the galaxy datatype formats for this galaxy output"""
        ...

class FormatFetchStrategy(FetchStrategy):
    def fetch(self, gxout: GxOutput, inputs: ParamRegister) -> list[str]:
        return str(gxout.format).split(',')

class DefaultFormatFetchStrategy(FetchStrategy):
    def fetch(self, gxout: GxOutput, inputs: ParamRegister) -> list[str]:
        return str(gxout.default_format).split(',')

class FormatSourceFetchStrategy(FetchStrategy):
    def fetch(self, gxout: GxOutput, inputs: ParamRegister) -> list[str]:
        param = inputs.get(gxout.format_source, strategy='lca')
        if param:
            return param.formats
        return []

class CollectorFetchStrategy(FetchStrategy):
    def fetch(self, gxout: GxOutput, inputs: ParamRegister) -> list[str]:
        coll = gxout.dataset_collector_descriptions[0]
        if coll.default_ext:
            return str(coll.default_ext).split(',')
        return []

class FallbackFetchStrategy(FetchStrategy):
    def fetch(self, gxout: GxOutput, inputs: ParamRegister) -> list[str]:
        return []

def select_fetcher(gxout: GxOutput) -> FetchStrategy:
    if has_format(gxout):
        return FormatFetchStrategy()
    elif has_default_format(gxout):
        return DefaultFormatFetchStrategy()
    elif has_format_source(gxout):
        return FormatSourceFetchStrategy()
    elif has_dataset_collector(gxout):
        return CollectorFetchStrategy()
    else:
        return FallbackFetchStrategy()














from abc import ABC, abstractmethod
from typing import Any


class Param(ABC):
    name: str
    formats: list[str] = []

    @property
    @abstractmethod
    def default(self) -> Any:
        ...

    @property
    @abstractmethod
    def docstring(self) -> str:
        ...
    
    @property
    @abstractmethod
    def optional(self) -> bool:
        ...
    
    @property
    @abstractmethod
    def array(self) -> bool:
        ...








from galaxy.tools.parameters.basic import ToolParameter as GalaxyParam
from galaxy2janis.runtime.exceptions import ParamNotSupportedError

from ..param.Param import Param
from ..param.InputParam import (
    InputParam,
    SelectOption,
    TextParam,
    IntegerParam,
    FloatParam,
    BoolParam,
    SelectParam,
    DataParam,
    DataCollectionParam
)


def parse_input_param(gxparam: GalaxyParam) -> Param:
    factory = InputParamFactory()
    return factory.produce(gxparam)

class InputParamFactory:
    def produce(self, gxparam: GalaxyParam) -> Param:
        match gxparam.type: # type: ignore
            case 'text':
                param = self.parse_text_param(gxparam)
            case 'integer':
                param = self.parse_int_param(gxparam)
            case 'float':
                param = self.parse_float_param(gxparam)
            case 'boolean':
                param = self.parse_bool_param(gxparam)
            case 'select':
                param = self.parse_select_param(gxparam)
            case 'data':
                param = self.parse_data_param(gxparam)
            case 'data_collection':
                param = self.parse_data_collection_param(gxparam)
            case 'data_column':
                param = self.parse_int_param(gxparam)
            case 'color':
                param = self.parse_text_param(gxparam)
                #logging.color_param_ignored()
            case _:
                raise ParamNotSupportedError(f'unknown param type: {str(gxparam.type)}')
        
        param = self.map_common_fields(gxparam, param)
        return param 
        
    def map_common_fields(self, gxparam: GalaxyParam, param: InputParam) -> InputParam:
        param.label = str(gxparam.label)
        param.helptext = str(gxparam.help)
        param.argument = gxparam.argument
        param.set_optionality(bool(gxparam.optional))
        return param

    def parse_text_param(self, gxparam: GalaxyParam) -> TextParam:
        param = TextParam(str(gxparam.flat_name))
        param.value = gxparam.value
        return param

    def parse_int_param(self, gxparam: GalaxyParam) -> IntegerParam:
        param = IntegerParam(str(gxparam.flat_name))
        if hasattr(gxparam, 'value'):
            param.value = gxparam.value
        if hasattr(gxparam, 'min'):
            param.min = gxparam.min
        if hasattr(gxparam, 'max'):
            param.max = gxparam.max
        return param

    def parse_float_param(self, gxparam: GalaxyParam) -> FloatParam:
        param = FloatParam(str(gxparam.flat_name))
        param.value = gxparam.value
        param.min = gxparam.min
        param.max = gxparam.max
        return param

    def parse_bool_param(self, gxparam: GalaxyParam) -> BoolParam:
        param = BoolParam(str(gxparam.flat_name))
        param.checked = bool(gxparam.checked)
        param.truevalue = str(gxparam.truevalue)
        param.falsevalue = str(gxparam.falsevalue)
        return param

    def parse_select_param(self, gxparam: GalaxyParam) -> SelectParam:
        # TODO this could be dynamic options!
        param = SelectParam(str(gxparam.flat_name))
        param.multiple = bool(gxparam.multiple)
        if hasattr(gxparam, 'static_options') and len(gxparam.static_options) > 0:
            for opt in gxparam.static_options:
                option = SelectOption(value=opt[1], selected=opt[2], ui_text=opt[0])
                param.options.append(option)
        return param

    def parse_data_param(self, gxparam: GalaxyParam) -> DataParam:
        param = DataParam(str(gxparam.flat_name))
        param.formats = gxparam.extensions
        param.multiple = bool(gxparam.multiple)
        return param

    def parse_data_collection_param(self, gxparam: GalaxyParam) -> DataCollectionParam:
        name = str(gxparam.flat_name)
        collection_type = gxparam.collection_types[0]
        param = DataCollectionParam(name, collection_type)
        param.formats = gxparam.extensions
        return param






from dataclasses import dataclass
from typing import Optional

from ..tokens import Token
from ..tokens import spawn_end_sentinel
from ..components import CommandComponent
from ..components import StreamMerge
from ..components import Tee

from .ComponentOrderingStrategy import SimplifiedComponentOrderingStrategy


@dataclass
class EPathPosition:
    ptr: int
    token: Token
    ignore: bool = False
    component: Optional[CommandComponent] = None


class ExecutionPath:
    id: int

    def __init__(self, tokens: list[Token]):
        self.positions: list[EPathPosition] = self._init_positions(tokens)
        self.tokens_to_excise: list[int] = []

    def get_components(self) -> list[CommandComponent]:
        components = self._get_component_list()
        ordering_strategy = SimplifiedComponentOrderingStrategy()
        return ordering_strategy.order(components)
    
    def _get_component_list(self) -> list[CommandComponent]:
        """
        gets the CommandComponents in this EPath (unique)
        preserves ordering
        """
        ignore = [Tee, StreamMerge]
        out: list[CommandComponent] = []
        for position in self.positions:
            if position.component and type(position.component) not in ignore and position.component not in out:
                out.append(position.component)
        return out
    
    def _init_positions(self, tokens: list[Token]) -> list[EPathPosition]:
        positions = [EPathPosition(i, token) for i, token in enumerate(tokens)]
        end_sentinel = spawn_end_sentinel()
        positions.append(EPathPosition(len(positions), end_sentinel))
        return positions

    def __str__(self) -> str:
        out: str = '\n'
        for position in self.positions:
            out += f'{str(position.ptr):<3}{position.token.text[:39]:<40}{position.token.ttype:<35}\n'
        return out






import os


def init_file(path: str, override: bool=False, contents: str='') -> None:
    _make_parent_folders(path)
    if override:
        _unsafe_init_file(path, contents)
    else:
        _safe_init_file(path, contents)

def init_folder(path: str, override: bool=False) -> None:
    _make_parent_folders(path)
    if override:
        _unsafe_init_folder(path)
    else:
        _safe_init_folder(path)

def _make_parent_folders(path: str) -> None:
    heirarchy = _gen_heirarchy(path)
    for folder in heirarchy[:-1]:
        _safe_init_folder(folder)

def _safe_init_file(path: str, contents: str) -> None:
    if not os.path.exists(path):
        with open(path, 'w') as fp:
            fp.write(contents)
    
def _unsafe_init_file(path: str, contents: str) -> None:
    if os.path.exists(path):
        os.remove(path)
    with open(path, 'w') as fp:
        fp.write(contents)

def _safe_init_folder(path: str) -> None:
    if not os.path.isdir(path):
        os.mkdir(path)

def _unsafe_init_folder(path: str) -> None:
    if os.path.isdir(path):
        os.remove(path)
    os.mkdir(path)

def _gen_heirarchy(path: str) -> list[str]:
    out: list[str] = []
    heirarchy = path.split('/')
    for i, elem in enumerate(heirarchy):
        parent = '/'.join(heirarchy[:i])
        if parent == '':
            out.append(elem)
        else:
            out.append(f'{parent}/{elem}')
    return out






from dataclasses import dataclass
from typing import Tuple
from galaxy2janis.gx.command.components import OutputComponent

from ..TextRender import TextRender
from ..tool.ToolOutputText import ToolOutputText
from .. import ordering


@dataclass
class ToolOutputSectionText(TextRender):
    def __init__(self, entity: list[OutputComponent]):
        super().__init__()
        self.entity = entity
    
    @property
    def imports(self) -> list[Tuple[str, str]]:
        imports: list[Tuple[str, str]] = []
        for out in self.entity:
            imports += ToolOutputText(out).imports
        imports = list(set(imports))
        return ordering.order_imports(imports)

    def render(self) -> str:
        out_str: str = ''
        
        out_str += 'outputs = [\n'
        for out in self.entity:
            render = ToolOutputText(out)
            out_str += f'\t{render.render()},\n'
        out_str += '\n]\n'

        return out_str




from typing import Optional, Tuple
import regex as re
import numpy as np

from .patterns import QUOTED_SECTION 


def get_matches(the_string: str, expression: str) -> list[re.Match[str]]:
    matches = re.finditer(expression, the_string)
    return [m for m in matches]

def get_next_word(word: str, delim: str, text: str) -> Optional[str]:
    NEXT_WORD = r'(?<=(?:\s|^))' + f'{word}{delim}' + r'+?([\w\d\'"${}\\_.\-\:/]+)(?=\s|$)'
    matches = re.finditer(NEXT_WORD, text)
    matches = [m for m in matches]
    if matches:
        value = matches[0].group(1)
        value = value.strip('"\'')
        return value
    return None

def get_preceeding_dashes(search_term: str, text: str) -> list[str]:
    PRECEEDING_DASHES = r'(?<![$.{])(-+?)' + fr'({search_term})' + r'(?=[\s=:]|$|[\'"])'
    matches = re.finditer(PRECEEDING_DASHES, text)
    return [m.group(1) for m in matches]

def get_quoted_sections(the_string: str):
    # find the areas of the string which are quoted
    matches = re.finditer(QUOTED_SECTION, the_string)
    quoted_sections = [(m.start(), m.end()) for m in matches]

    # transform to mask
    quotes_mask = np.zeros(len(the_string)) # type: ignore
    for start, end in quoted_sections:
        quotes_mask[start: end] = 1
    
    return quotes_mask

def find_unquoted(the_string: str, pattern: str) -> Tuple[int, int]:
    """
    finds the pattern in string. ensures section is not quoted. 
    """
    # find quoted sections of input string
    quotes_mask = get_quoted_sections(the_string)
    
    # get pattern match locations
    matches = re.finditer(pattern, the_string)
    match_spans = [(m.start(), m.end()) for m in matches]

    # check each match to see if its in a quoted section
    if len(match_spans) > 0:
        for start, end in match_spans:
            if sum(quotes_mask[start: end]) == 0:
                # return position of first unquoted match
                return start, end
    return -1, -1

# def get_unpaired_quotes_start(the_string: str) -> int:
#     # find all quotes
#     matches = re.finditer(QUOTES, the_string)
#     quote_locations = [m.start() for m in matches]
#     # find paired quotes sections
#     quotes_mask = get_quoted_sections(the_string)
#     # check each quote is part of quoted section
#     for loc in quote_locations:
#         if quotes_mask[loc] != 1:
#             return loc
#     return -1





from abc import ABC, abstractmethod
from typing import Any
from galaxy2janis import expressions
import json

from .JanisDatatype import JanisDatatype
from .conversion import galaxy_to_janis
from . import core



### SHARED FUNCTIONS ###

def types_from_param(entity: Any) -> list[str]:
    if entity.gxparam and entity.gxparam.formats:
        return entity.gxparam.formats
    return []

def types_from_values(entity: Any) -> list[str]:
    values = entity.values.unique
    if all([expressions.is_script(v) for v in values]):
        return ['file']
    elif all([expressions.is_int(v) for v in values]):
        return ['integer']
    elif all([expressions.is_float(v) for v in values]):
        return ['float']
    elif all([expressions.is_var(v) for v in values]):
        return ['string']
    return []

def types_from_default(entity: Any) -> list[str]:
    default = entity.default_value
    if default:
        if expressions.is_int(default):
            return ['integer']
        if expressions.is_float(default):
            return ['float']
        if expressions.is_var(default) or expressions.has_var(default):
            return ['string']
    return []

def types_from_extension(output: Any) -> list[str]:
    if output.gxparam and hasattr(output.gxparam, 'discover_pattern'):
        pattern: str = output.gxparam.discover_pattern
        if pattern and '.' in pattern:
            compressed_types = ['gz', 'bz2']
            p_split = pattern.split('.')
            while p_split[-1] in compressed_types:
                p_split = p_split[:-1]
            ext = p_split[-1]
            return [f'.{ext}']
    return []


### STRATEGIES ###

class DatatypeGetStrategy(ABC):
    @abstractmethod
    def get(self, entity: Any) -> list[JanisDatatype]:
        """parses an entity to return the janis datatype(s)"""
        ...

class PositionalStrategy(DatatypeGetStrategy):
    def get(self, entity: Any) -> list[JanisDatatype]:
        gxtypes = types_from_param(entity)
        if not gxtypes:
            gxtypes = types_from_values(entity)
        if not gxtypes:
            gxtypes = types_from_default(entity)
        if not gxtypes:
            gxtypes = ['string']
        return galaxy_to_janis(gxtypes)

class FlagStrategy(DatatypeGetStrategy):
    def get(self, entity: Any) -> list[JanisDatatype]:
        return [core.bool_t]

class OptionStrategy(DatatypeGetStrategy):
    def get(self, entity: Any) -> list[JanisDatatype]:
        gxtypes = types_from_param(entity)
        if not gxtypes:
            gxtypes = types_from_values(entity)
        if not gxtypes:
            gxtypes = types_from_default(entity)
        if not gxtypes:
            gxtypes = ['string']
        return galaxy_to_janis(gxtypes)

class OutputStrategy(DatatypeGetStrategy):
    def get(self, entity: Any) -> list[JanisDatatype]:
        gxtypes = types_from_param(entity)
        if not gxtypes:
            gxtypes = types_from_extension(entity)
        if not gxtypes:
            gxtypes = ['file']
        return galaxy_to_janis(gxtypes)

class WorkflowInputStrategy(DatatypeGetStrategy):
    def get(self, entity: Any) -> list[JanisDatatype]:
        return [entity.datatype]

class GalaxyInputStepStrategy(DatatypeGetStrategy):
    def get(self, entity: dict[str, Any]) -> list[JanisDatatype]:
        tool_state = json.loads(entity['tool_state'])
        if 'format' in tool_state:
            return galaxy_to_janis(tool_state['format'])
        else:
            return [core.file_t]



# these are the entities which we can get a datatype for

strategy_map = {
    'Positional': PositionalStrategy,
    'Flag': FlagStrategy,
    'Option': OptionStrategy,
    'RedirectOutput': OutputStrategy,
    'InputOutput': OutputStrategy,
    'WildcardOutput': OutputStrategy,
    'GalaxyInputStep': GalaxyInputStepStrategy,
    'WorkflowInput': WorkflowInputStrategy,
}





        





from typing import Any, Optional
from galaxy2janis import expressions

from ...command.components.outputs.OutputComponent import OutputComponent
from ...command.components import InputComponent


def get_comptype(component: InputComponent | OutputComponent) -> str:
    return type(component).__name__.lower() 

def select_input_value_type(component: Optional[InputComponent], value: Any) -> str:
    """
    only StaticValueLinkingStrategy and DefaultValueLinkingStrategy 
    call select_input_value_type(). don't need to worry about CONNECTION and RUNTIME_VALUE
    """
    if is_bool(value):
        return 'boolean'
    elif component and is_numeric(component, value):
        return 'numeric'
    elif is_none(value):
        return 'none'
    elif expressions.is_var(value) or expressions.has_var(value):
        return 'env_var'
    else:
        return 'string'

def is_bool(value: Any) -> bool:
    if isinstance(value, bool):
        return True
    return False

def is_none(value: Any) -> bool:
    if value is None:
        return True
    return False

def is_numeric(component: InputComponent | OutputComponent, value: Any) -> bool:
    if _has_numeric_datatype(component):
        if expressions.is_int(str(value)) or expressions.is_float(str(value)):
            return True
    return False

def _has_numeric_datatype(component: InputComponent | OutputComponent) -> bool:
    jtype = component.datatype
    numeric_classes = ['Int', 'Float']
    if jtype.classname in numeric_classes:
        return True
    return False
    







from typing import Any

from galaxy2janis.entities.workflow import StepOutput
from galaxy2janis.entities.workflow import WorkflowStep
from galaxy2janis.entities.workflow import Workflow

from galaxy2janis import mapping
from galaxy2janis import settings


def ingest_workflow_steps_outputs(janis: Workflow, galaxy: dict[str, Any]) -> None:
    ingestor = OutputParser(janis, galaxy)
    ingestor.parse()


class OutputParser:
    def __init__(self, janis: Workflow, galaxy: dict[str, Any]):
        self.janis = janis
        self.galaxy = galaxy
    
    def parse(self) -> None:
        for step in self.galaxy['steps'].values():
            if step['type'] == 'tool':
                self.parse_step_outputs(step)

    def parse_step_outputs(self, galaxy_step: dict[str, Any]) -> None:
        janis_step = mapping.step(galaxy_step['id'], self.janis, self.galaxy)
        settings.tool.set(wrapper=janis_step.metadata.wrapper)
        for galaxy_out in galaxy_step['outputs']:
            janis_out = init_tool_step_output(janis_step, galaxy_step, galaxy_out)
            galaxy_out['janis_uuid'] = janis_out.uuid # entity linking
            janis_step.outputs.add(janis_out)


def init_tool_step_output(janis_step: WorkflowStep, step: dict[str, Any], output: dict[str, Any]) -> StepOutput:
    tool_output = mapping.tool_output(output['name'], janis_step.tool)
    assert(tool_output)
    return StepOutput(
        step_uuid=janis_step.uuid,
        is_wflow_out=is_workflow_output(step, output),
        tool_output=tool_output,
    )

def is_workflow_output(step: dict[str, Any], output: dict[str, Any]) -> bool:
    if step['workflow_outputs']:
        for details in step['workflow_outputs']:
            if details['output_name'] == output['name']:
                return True
    elif not step['workflow_outputs']:
        return True
    return False






from dataclasses import dataclass
from typing import Optional, Tuple
from galaxy2janis.utils.general import global_align

from .requirements import ContainerRequirement, CondaRequirement
from .citations import Citation


Requirement = ContainerRequirement | CondaRequirement

@dataclass
class ToolXMLMetadata:
    name: str
    id: str
    version: str
    description: str
    help: str
    requirements: list[Requirement]
    citations: list[Citation]
    creator: Optional[str] = None
    # TODO
    url: Optional[str] = None
    owner: Optional[str] = None

    def set_url(self, owner: str) -> None:
        self.owner = owner

    def set_owner(self, owner: str) -> None:
        self.owner = owner

    def get_main_requirement(self) -> Requirement:
        """
        sets the main tool requirement
        if no requirements parsed, sets to tool info
        else, sets from the parsed requirements. 
        method:
            - align each requirement to the tool id
            - pick the one with best alignment score
        """
        if len(self.requirements) == 0:
            return CondaRequirement(self.id, self.version)
        
        similarity_scores = self.get_req_similarity_scores()
        similarity_scores.sort(key=lambda x: x[1], reverse=True)
        main_requirement: Requirement = similarity_scores[0][0]
        return main_requirement

    def get_req_similarity_scores(self) -> list[Tuple[Requirement, float]]:
        scores: list[Tuple[Requirement, float]] = []
        for req in self.requirements:
            similarity = global_align(req.name, self.id)
            scores.append((req, similarity))
        return scores
             
    def get_main_citation(self) -> str:
        biotools_citations = [x for x in self.citations if x.type == 'biotools']
        doi_citations = [x for x in self.citations if x.type == 'doi']
        bibtex_citations = [x for x in self.citations if x.type == 'bibtex']
        if biotools_citations:
            return biotools_citations[0].text
        elif doi_citations:
            return doi_citations[0].text
        elif bibtex_citations:
            return bibtex_citations[0].text
        return 'tool xml missing citation'

    def get_doi_citation(self) -> Optional[str]:
        doi_citations = [x for x in self.citations if x.type == 'doi']
        if len(doi_citations) > 0:
            return doi_citations[0].text
        return None






from dataclasses import dataclass
from functools import cached_property
from typing import Optional, Tuple

from galaxy2janis.entities.workflow import Workflow
from galaxy2janis.entities.workflow import WorkflowStep
from galaxy2janis.entities.workflow import InputValue, ConnectionInputValue, StaticInputValue, WorkflowInputInputValue
from galaxy2janis.gx.command.components import Positional

from .WorkflowInputText import WorkflowInputText
from ..TextRender import TextRender
from .. import formatting
from .. import ordering


### HELPER METHODS ### 

def note() -> str:
    return"""\"\"\"
FOREWORD ----------
Please see [WEBLINK] for information on RUNTIME VALUES
Please see [WEBLINK] for information on the PRE TASK, TOOL STEP, POST TASK structure
\"\"\"
"""

### HELPER CLASSES ###

@dataclass
class ToolInputLine:
    tag_and_value: str 
    special_label: str
    prefix_label: str
    default_label: str
    datatype_label: str

    @cached_property
    # NOTE: I dont like overriding __len__ or any other dunders
    def length(self) -> int:  
        return len(f'{self.tag_and_value},')

    def render(self, info_indent: int) -> str:
        left = f'{self.tag_and_value},'
        right = f'#'
        if self.prefix_label != '':
            right += f' {self.prefix_label}'
        if self.datatype_label != '':
            right += f' [{self.datatype_label}]'
        if self.special_label != '':
            right += f' ({self.special_label})'
        if self.default_label != '':
            right += f' ({self.default_label})'
        justified = f'\t\t{left:<{info_indent+2}}{right}'
        return justified


class ToolInputLineFactory:
    unknown_count = 0

    def create(self, invalue: InputValue) -> ToolInputLine:
        if not invalue.component:
            self.unknown_count += 1
            tag_and_value = f'#{invalue.input_tag}{self.unknown_count}={invalue.input_value}'
        else:
            tag_and_value = f'{invalue.input_tag}={invalue.input_value}'
        return ToolInputLine(
            tag_and_value=tag_and_value,
            special_label=self.get_special_label(invalue),
            prefix_label=self.get_prefix_label(invalue),
            default_label=self.get_default_label(invalue),
            datatype_label=self.get_datatype_label(invalue),
        )

    def get_special_label(self, invalue: InputValue) -> str:
        if isinstance(invalue, WorkflowInputInputValue):
            return 'WORKFLOW INPUT'
            # if invalue.is_runtime:
            #     return 'RUNTIME VALUE'
            # else:
            #     return 'WORKFLOW INPUT'
        if isinstance(invalue, ConnectionInputValue):
            return 'CONNECTION'
        return ''
    
    def get_prefix_label(self, invalue: InputValue) -> str:
        if invalue.component:
            if hasattr(invalue.component, 'prefix'):
                return invalue.component.prefix  # type: ignore
            elif isinstance(invalue, Positional):
                return f'position {invalue.cmd_pos}'
        return ''
    
    def get_default_label(self, invalue: InputValue) -> str:
        if isinstance(invalue, StaticInputValue):
            if invalue.default:
                return 'DEFAULT'
        return ''
    
    def get_datatype_label(self, invalue: InputValue) -> str:
        if invalue.component:
            return formatting.format_typestr(invalue.component, fmt='value')
        return ''


### MAIN CLASS ###

class StepText(TextRender):
    def __init__(
        self, 
        entity: WorkflowStep, 
        janis: Workflow,
        render_note: bool=False,
        render_imports: bool=False,
        render_runtime_inputs: bool=False
    ):
        super().__init__()
        self.entity = entity
        self.janis = janis
        self.render_note = render_note
        self.render_imports = render_imports
        self.render_runtime_inputs = render_runtime_inputs
    
    @cached_property
    def lines(self) -> list[ToolInputLine]:
        invalues = self.entity.inputs.all
        invalues = ordering.order_step_inputs(invalues)
        factory = ToolInputLineFactory()
        return [factory.create(val) for val in invalues]
    
    @cached_property
    def line_len(self) -> int:
        return max([line.length for line in self.lines])

    @property
    def imports(self) -> list[Tuple[str, str]]:
        imports: list[Tuple[str, str]] = []
        if len(self.get_scatter_input_names()) > 1:
            imports.append(('janis_core', 'ScatterDescription'))
            imports.append(('janis_core', 'ScatterMethods'))
        return imports

    def render(self) -> str:
        out_str: str = ''
        if self.render_note:
            out_str += f'{note()}\n'
        if self.render_imports:
            out_str += f'{formatting.format_imports(self.imports)}\n'
        if self.render_runtime_inputs:
            out_str += f'{self.format_runtime_inputs()}\n'
        
        out_str += f'{self.format_tool()}\n'
        return out_str

    def format_runtime_inputs(self) -> str:
        inputs = self.entity.inputs.all
        inputs = [x for x in inputs if isinstance(x, WorkflowInputInputValue)]
        inputs = [x for x in inputs if x.is_runtime]
        winps = [self.janis.get_input(x.input_uuid) for x in inputs]
        out_str = ''
        for winp in winps:
            out_str += f'{WorkflowInputText(winp).render()}\n'
        return out_str

    def format_tool(self) -> str:
        step_tag = self.entity.tag
        tool_tag = self.entity.tool.tag
        scatter_stmt = self.format_scatter()
        
        out_str: str = ''
        out_str += 'w.step(\n'
        out_str += f'\t"{step_tag}",\n'
        out_str += f'\t{tool_tag}(\n'
        for line in self.lines:
            out_str += f'{line.render(self.line_len)}\n'
        out_str += '\t)'
        if scatter_stmt:
            out_str += f',\n{scatter_stmt}'
        out_str += '\n)'

        return out_str

    def format_scatter(self) -> Optional[str]:
        names = self.get_scatter_input_names()
        return self.format_scatter_statement(names)

    def get_scatter_input_names(self) -> list[str]:
        # get scatter inputs names
        unknown_count: int = 0
        targets: list[str] = []
        invalues = self.entity.inputs.all
        invalues = ordering.order_step_inputs(invalues)
        for val in invalues:
            if val.scatter:
                if not val.component:
                    unknown_count += 1
                    targets.append(f'{val.input_tag}{unknown_count}')
                else:
                    targets.append(val.input_tag)
        return targets

    def format_scatter_statement(self, names: list[str]) -> Optional[str]:
        # format text
        if len(names) == 0:
            return None
        elif len(names) == 1:
            return f'\tscatter="{names[0]}"'
        else:
            return f'\tscatter=ScatterDescription(fields={repr(names)}, method=ScatterMethods.dot)'


        






from dataclasses import dataclass
from typing import Any, Optional

from .Param import Param


class InputParam(Param):
    def __init__(self, name: str):
        self.name: str = name
        self.formats: list[str] = []
        self.label: str = ''
        self.argument: Optional[str] = None
        self.helptext: str = ''
        self._optional: bool = False

    def set_optionality(self, value: bool) -> None:
        self._optional = value

    @property
    def default(self) -> Any:
        return None

    @property
    def docstring(self) -> str:
        return self.generic_get_docstring()

    def generic_get_docstring(self) -> str:
        if self.helptext and self.label:
            return self.label + '. ' + self.helptext
        elif self.helptext:
            return self.helptext
        elif self.label:
            return self.label
        return ''
    
    @property
    def optional(self) -> bool:
        return self._optional
    
    @property
    def array(self) -> bool:
        return False



class TextParam(InputParam):
    def __init__(self, name: str):
        super().__init__(name)
        self.value: Optional[str] = None
        self.formats: list[str] = ['string']

    @property
    def default(self) -> Any:
        if self.value:
            return self.value
        return None
    

class IntegerParam(InputParam):
    def __init__(self, name: str):
        super().__init__(name)
        self.value: Optional[str] = None
        self.min: Optional[str] = None
        self.max: Optional[str] = None
        self.formats: list[str] = ['integer']

    @property
    def default(self) -> Any:
        if self.value:
            return self.value
        # elif self.min:
        #     return self.min
        # elif self.max:
        #     return self.max
        return None


# YES I KNOW THIS IS ESSENTIALLY DUPLICATED FROM INTEGERParam SUE ME
class FloatParam(InputParam):
    def __init__(self, name: str):
        super().__init__(name)
        self.value: Optional[str] = None
        self.min: Optional[str] = None
        self.max: Optional[str] = None
        self.formats: list[str] = ['float']

    @property
    def default(self) -> Any:
        if self.value:
            return self.value
        elif self.min:
            return self.min
        elif self.max:
            return self.max
        return None


class BoolParam(InputParam):
    def __init__(self, name: str):
        super().__init__(name)
        self.checked: bool = True
        self.truevalue: str = ''
        self.falsevalue: str = ''
        self.formats: list[str] = ['boolean']

    @property
    def default(self) -> str:
        if self.checked:
            return self.truevalue
        return self.falsevalue

    @property
    def docstring(self) -> str:
        docstring = self.generic_get_docstring()
        bool_values = [self.truevalue, self.falsevalue]
        bool_values = [v for v in bool_values if v != '']
        bool_values.sort()
        bool_str = ', '.join(bool_values)
        return f'{docstring}. possible values: {bool_str}'
    
    @property
    def optional(self) -> bool:
        return True

    def get_all_values(self, nonempty: bool=False) -> list[str]:
        values = [self.truevalue, self.falsevalue]
        empty_values = ['', 'none', 'None', 'null']
        if nonempty:
            values = [v for v in values if v not in empty_values]
        return values


@dataclass
class SelectOption:
    value: str
    selected: bool
    ui_text: str

class SelectParam(InputParam):
    def __init__(self, name: str):
        super().__init__(name)
        self.options: list[SelectOption] = []
        self.multiple: bool = False
        self.formats: list[str] = []

    @property
    def default(self) -> Any:
        for option in self.options:
            if option.selected:
                return option.value
        if len(self.options) > 0:
            return self.options[0].value
        return ''        

    @property
    def docstring(self) -> str:
        docstring = self.generic_get_docstring()
        option_values = [v.value for v in self.options]
        option_values.sort()
        option_str = ', '.join(option_values[:5])
        return f'{docstring}. possible values: {option_str}'
    
    @property
    def optional(self) -> bool:
        if self.multiple or self._optional:
            return True
        return False
    
    @property
    def array(self) -> bool:
        if self.multiple:
            return True
        return False

    def get_all_values(self, nonempty: bool=False) -> list[str]:
        values = [opt.value for opt in self.options]
        empty_values = ['', 'none', 'None', 'null']
        if nonempty:
            values = [v for v in values if v not in empty_values]
        return values
    

class DataParam(InputParam):
    def __init__(self, name: str):
        super().__init__(name)
        self.formats: list[str] = []
        self.multiple: bool = False
    
    @property
    def array(self) -> bool:
        return self.multiple


class DataCollectionParam(InputParam):
    def __init__(self, name: str, collection_type: str):
        super().__init__(name)
        self.collection_type = collection_type
        self.formats: list[str] = []
    
    @property
    def array(self) -> bool:
        return True







from dataclasses import dataclass

from ..tokens import Token
from .DynamicCommandStatement import DynamicCommandStatement


@dataclass
class CommandString:
    main: DynamicCommandStatement
    preprocessing: list[DynamicCommandStatement]
    postprocessing: list[DynamicCommandStatement]

    def get_original_tokens(self) -> list[Token]:
        """gets the original tokenized form of the command string"""
        out: list[Token] = []
        for statement in self.preprocessing:
            out += statement.get_tokens()
        out += self.main.get_tokens()
        for statement in self.postprocessing:
            out += statement.get_tokens()
        return out








from typing import Any

from .blocks import get_next_block 
from .blocks import CheetahBlock
from .. import utils


def sectional_evaluate(text: str, inputs: dict[str, Any]) -> str:
    raw_lines = utils.split_lines_blanklines(text)
    evaluator = PartialCheetahEvaluator(raw_lines, inputs)
    eval_lines = evaluator.evaluate()
    return utils.join_lines(eval_lines)


class PartialCheetahEvaluator:
    def __init__(self, lines: list[str], input_dict: dict[str, Any]):
        self.lines = lines
        self.input_dict = input_dict
        self.ptr: int = 0

    def evaluate(self) -> list[str]:
        try:
            return self.evaluation_worker()
        except Exception as e:
            return self.lines

    def evaluation_worker(self) -> list[str]:
        while self.ptr < len(self.lines):
            block = get_next_block(self.ptr, self.lines)
            block.evaluate(self.input_dict)
            self.update_lines(block)
            self.update_ptr(block)
        return self.lines

    def update_lines(self, block: CheetahBlock) -> None:
        if block.evaluated:
            old_lines_top = self.lines[:block.start]
            old_lines_bottom = self.lines[block.stop + 1:]
            self.lines = old_lines_top + block.lines + old_lines_bottom
    
    def update_ptr(self, block: CheetahBlock) -> None:
        if block.evaluated:
            self.ptr += 1  # success, go to next line
        else:
            self.ptr += block.height # TODO CHECK









from typing import Type

from ...gxtool.XMLToolDefinition import XMLToolDefinition
from ..tokens import TokenType
from ..Command import Command
from .ExecutionPath import ExecutionPath
from .Annotator import (
    Annotator, 
    CompoundOptionAnnotator, 
    FlagAnnotator, 
    OptionAnnotator, 
    PositionalAnnotator, 
    RedirectAnnotator, 
    StreamMergeAnnotator, 
    TeeAnnotator
)


"""
iterates through a ExecutionPath, yielding the current tokens being assessed.
keeps track of the location we are in the ExecutionPath

Example ExecutionPath:
abricate $sample_name --no-header --minid=80 --db = card

0            1             2            3           4          5        6           7          8
abricate     $sample_name  --no-header  --minid     =          80       --db        =          card
RAW_STRING   ENV_VAR       RAW_STRING   RAW_STRING  KV_LINKER  RAW_INT  RAW_STRING  KV_LINKER  RAW_STRING
positional1  positional2   flag1        option1     option1    option1  option2     option2    option2

at token 2, we would return {
    ctoken: token 2  (--no-header's token),
    ntokens: []
}
at token 3, we would return {
    ctoken: token 3 (--minid's token),
    ntokens: [token 4, token 5]
}

"""

linux_constructs: list[Type[Annotator]] = [
    StreamMergeAnnotator,  
    RedirectAnnotator,  
    TeeAnnotator,  
]

# annotation order matters ?
tool_arguments: list[Type[Annotator]] = [
    OptionAnnotator,   # priority 1?
    CompoundOptionAnnotator,
    FlagAnnotator,
    PositionalAnnotator
]


class GreedyExecutionPathAnnotator:
    def __init__(self, epath: ExecutionPath, xmltool: XMLToolDefinition, command: Command):
        self.epath = epath 
        self.xmltool = xmltool
        self.command = command

    def annotate(self) -> ExecutionPath:
        self.mark_ignore_tokens()
        self.annotate_via_param_args()
        self.annotate_linux_constructs()
        self.annotate_tool_arguments()
        self.transfer_gxparams()
        return self.epath

    def mark_ignore_tokens(self) -> None:
        ignore_tokens = [
            TokenType.FUNCTION_CALL,
            TokenType.BACKTICK_SHELL_STATEMENT,
        ]
        for position in self.epath.positions:
            if position.token.ttype in ignore_tokens:
                position.ignore = True
        
    def annotate_via_param_args(self) -> None:
        arguments: set[str] = set([param.argument for param in self.xmltool.inputs.list() if param.argument]) # type: ignore
        ptr = 0
        while ptr < len(self.epath.positions) - 1:
            token = self.epath.positions[ptr].token
            if token.text in arguments:
                token.ttype = TokenType.FORCED_PREFIX
                ptr = self.annotate_position(ptr, annotators=tool_arguments)
            else:
                ptr += 1
    
    def annotate_linux_constructs(self) -> None:
        self.iter_annotate(annotators=linux_constructs)
    
    def annotate_tool_arguments(self) -> None:
        self.iter_annotate(annotators=tool_arguments)

    def iter_annotate(self, annotators: list[Type[Annotator]]) -> None:
        ptr = 0 # reset
        while ptr < len(self.epath.positions) - 1:
            position = self.epath.positions[ptr]
            if not position.ignore:
                ptr = self.annotate_position(ptr, annotators=annotators)
            else:
                ptr += 1

    def annotate_position(self, ptr: int, annotators: list[Type[Annotator]]) -> int:
        for annotator in annotators:
            a = annotator(ptr, self.epath.positions)
            a.annotate()
            if a.success:
                return a.calculate_next_ptr_pos()
        return ptr + 1

    def transfer_gxparams(self) -> None:
        for position in self.epath.positions:
            if position.component and position.token.gxparam:
                position.component.gxparam = position.token.gxparam





# refgenconf needs the following aliases to be done manually.
# setting these aliases here is global scope. 
# this allows refgenconf==0.9.3 to be used in python 3.10.
import collections.abc
collections.Iterable = collections.abc.Iterable
collections.Mapping = collections.abc.Mapping





from galaxy2janis.gx.gxtool.requirements import ContainerRequirement

from ..Container import Container
from .Fetcher import Fetcher


class ContainerReqFetcher(Fetcher):
    
    def fetch(self, requirement: ContainerRequirement) -> list[Container]:
        info: dict[str, str] = {
            'image_type': requirement.image_type,
            'repo': requirement.name,
            'tag': requirement.version,
            'url': requirement.url,
            '_timestamp': 'Tue, 14 June 1994 23:56:01 -0000', # my bday lol
        }
        return [Container(info)]





from galaxy2janis import aliases
from galaxy2janis.logs import logging

from galaxy2janis import paths
from galaxy2janis import settings

from galaxy2janis.fileio import write_tool
from galaxy2janis.startup import general_setup
from galaxy2janis.cli import CLIparser
from galaxy2janis.tool_mode import tool_mode
from galaxy2janis.workflow_mode import workflow_mode

import sys
from typing import Optional


"""
gxtool2janis program entry point
parses cli settings then hands execution to other files based on command
"""

def main():
    logging.configure_warnings()
    args = load_args()
    general_setup(args)
    run_sub_program(args)
    

def load_args() -> dict[str, Optional[str]]:
    cli = CLIparser(sys.argv)
    return cli.args

def run_sub_program(args: dict[str, Optional[str]]) -> None:
    match args['command']:
        case 'tool':
            run_tool_mode(args)
            #try_run_tool_mode(args)
        case 'workflow':
            run_workflow_mode(args)
            #try_run_workflow_mode(args)
        case _:
            pass

def run_tool_mode(args: dict[str, Optional[str]]):
    settings.tool.set(args)
    tool = tool_mode()
    path = paths.tool(tool.metadata.id)
    write_tool(tool, path=path)  # I dont like this design, but it may be necessary

def run_workflow_mode(args: dict[str, Optional[str]]):
    workflow_mode(args)


# for bulk parsing stat runs

def try_run_tool_mode(args: dict[str, Optional[str]]):
    try: 
        run_tool_mode(args)
    except Exception as e:
        # print('\n####################')
        # print(settings.tool.tool_id.upper())
        # print('####################\n')
        # print()
        print(e)
        logging.tool_exception()

def try_run_workflow_mode(args: dict[str, Optional[str]]):
    try: 
        run_workflow_mode(args)
    except Exception as e:
        print(e)
        logging.workflow_exception()
    

if __name__ == '__main__':
    main()







from typing import Optional
from dataclasses import dataclass


@dataclass
class JanisDatatype:
    format: str
    source: str
    classname: str
    extensions: Optional[str]
    import_path: str




import unittest
import datatypes
from datatypes.core import file_t, float_t, bool_t

# mock objects
from mock.mock_components import MOCK_POSIT1
from mock.mock_components import MOCK_FLAG1
from mock.mock_components import MOCK_OPTION1
from mock.mock_components import MOCK_REDIRECT1
from mock.mock_entities import MOCK_WORKFLOW_INPUT1


class TestDatatypeInference(unittest.TestCase):
    """
    tests the datatype which is assigned to an entity
    """

    def test_positional(self) -> None:
        self.assertEquals(datatypes.get(MOCK_POSIT1), file_t)
    
    def test_flag(self) -> None:
        self.assertEquals(datatypes.get(MOCK_FLAG1), bool_t)
    
    def test_option(self) -> None:
        self.assertEquals(datatypes.get(MOCK_OPTION1), float_t)
    
    def test_outputs(self) -> None:
        self.assertEquals(datatypes.get(MOCK_REDIRECT1), file_t)
    
    def test_workflow_input(self) -> None:
        self.assertEquals(datatypes.get(MOCK_WORKFLOW_INPUT1), file_t)
    
    # def test_option_typestring(self) -> None:
    #     raise NotImplementedError()









import operator
from typing import Callable

from galaxy2janis.runtime.exceptions import AttributeNotSupportedError
from janis_core.tool.test_classes import (
    TTestExpectedOutput,
    TTestPreprocessor
)

from .checks import ValidCheck
from . import custom_ops


def unsupported_strategy(vcheck: ValidCheck) -> TTestExpectedOutput:
    raise AttributeNotSupportedError()

def not_implemented_strategy(vcheck: ValidCheck) -> TTestExpectedOutput:
    raise NotImplementedError(f'no test strategy for {vcheck.ctype}')

def lines_diff_strategy(vcheck: ValidCheck) -> TTestExpectedOutput:
    return TTestExpectedOutput(
        tag=f'{vcheck.outname}_{vcheck.ctype}',
        preprocessor=TTestPreprocessor.LinesDiff,
        operator=operator.eq,
        file_diff_source=vcheck.reffile,
        expected_value=int(vcheck.value),
    )

def has_size_strategy(vcheck: ValidCheck) -> TTestExpectedOutput:
    if vcheck.reffile:
        return TTestExpectedOutput(
            tag=f'{vcheck.outname}_{vcheck.ctype}',
            preprocessor=TTestPreprocessor.FileSize,
            operator=operator.eq,
            expected_file=str(vcheck.reffile)
        )
    elif vcheck.value:
        return TTestExpectedOutput(
            tag=f'{vcheck.outname}_{vcheck.ctype}',
            preprocessor=TTestPreprocessor.FileSize,
            operator=operator.eq,
            expected_value=int(vcheck.value)
        )

def has_text_strategy(vcheck: ValidCheck) -> TTestExpectedOutput:
    if vcheck.reffile:
        return TTestExpectedOutput(
            tag=f'{vcheck.outname}_{vcheck.ctype}',
            preprocessor=TTestPreprocessor.FileContent,
            operator=custom_ops.has_text,
            expected_file=str(vcheck.reffile)
        )
    elif vcheck.value:
        return TTestExpectedOutput(
            tag=f'{vcheck.outname}_{vcheck.ctype}',
            preprocessor=TTestPreprocessor.FileContent,
            operator=custom_ops.has_text,
            expected_value=str(vcheck.value['text'])
        )
    raise AttributeError('either value or reffile needs to be supplied to each ValidCheck')
    
def has_n_lines_strategy(vcheck: ValidCheck) -> TTestExpectedOutput:
    return TTestExpectedOutput(
        tag=f'{vcheck.outname}_{vcheck.ctype}',
        preprocessor=TTestPreprocessor.LineCount,
        operator=operator.eq,
        expected_value=int(vcheck.value['n'])
    )

def has_n_columns_strategy(vcheck: ValidCheck) -> TTestExpectedOutput:
    return TTestExpectedOutput(
        tag=f'{vcheck.outname}_{vcheck.ctype}',
        preprocessor=TTestPreprocessor.FileContent,
        operator=custom_ops.has_n_columns,
        expected_value=int(vcheck.value['n'])
    )

def has_text_matching_strategy(vcheck: ValidCheck) -> TTestExpectedOutput:
    return TTestExpectedOutput(
        tag=f'{vcheck.outname}_{vcheck.ctype}',
        preprocessor=TTestPreprocessor.FileContent,
        operator=custom_ops.has_text_matching,
        expected_value=str(vcheck.value['expression'])
    )

def ftype_strategy(vcheck: ValidCheck) -> TTestExpectedOutput:
    # return TTestExpectedOutput(
    #     tag=f'{vcheck.outname}_{vcheck.ctype}',
    #     preprocessor=custom_ftype_preprocessor,
    #     operator=in_list_op,
    #     #operator=operator.eq,
    #     file_diff_source=vcheck.reffile,
    #     expected_value=str(vcheck.value),
    # )
    raise AttributeNotSupportedError('cannot check filetype')




strategy_map: dict[str, Callable[..., TTestExpectedOutput]] = {
    'lines_diff': lines_diff_strategy,
    'ftype': ftype_strategy,
    'has_size': has_size_strategy,
    'has_text': has_text_strategy,
    'has_n_lines': has_n_lines_strategy,
    'has_n_columns': has_n_columns_strategy,
    'has_text_matching': has_text_matching_strategy,

    # not Implemented
    'not_has_text': not_implemented_strategy,
    'has_line': not_implemented_strategy,
    'has_line_matching': not_implemented_strategy,

    # not v0.1
    're_match': unsupported_strategy,
    're_match_multiline': unsupported_strategy,
    'has_archive_member': unsupported_strategy,
    'is_valid_xml': unsupported_strategy,
    'xml_element': unsupported_strategy,
    'has_element_with_path': unsupported_strategy,
    'has_n_elements_with_path': unsupported_strategy,
    'element_text_matches': unsupported_strategy,
    'element_text_is': unsupported_strategy,
    'attribute_matches': unsupported_strategy,
    'attribute_is': unsupported_strategy,
    'element_text': unsupported_strategy,
    'has_h5_keys': unsupported_strategy,
    'has_h5_attribute': unsupported_strategy
}


def map_ttestout(vcheck: ValidCheck) -> TTestExpectedOutput:
    map_func = strategy_map[vcheck.ctype]
    return map_func(vcheck)




from galaxy2janis import settings

from typing import Any
from galaxy2janis.runtime.exceptions import InputError
from galaxy2janis.utils import galaxy as utils


def workflow_setup(args: dict[str, Any]) -> None:
    update_workflow_settings(args)
    validate_workflow_settings()

def update_workflow_settings(args: dict[str, Any]) -> None:
    settings.workflow.set_path(args['infile'])
    settings.workflow.set_dev_partial_eval(args['dev_partial_eval'])


### VALIDATION ###

def validate_workflow_settings() -> None:
    if not _valid_workflow():
        raise InputError('please check workflow file path')

def _valid_workflow() -> bool:
    path = settings.workflow.workflow_path
    if utils.is_galaxy_workflow(path):
        return True
    return False







from typing import Any


from galaxy.tool_util.verify.asserts.text import (
    assert_has_line,
    assert_has_line_matching,
    assert_has_text,
    assert_has_text_matching,
    assert_not_has_text
)
from galaxy.tool_util.verify.asserts.tabular import (
    assert_has_n_columns
)


"""
# Implemented
'lines_diff': lines_diff_strategy,
'has_size': has_size_strategy,
'has_text': has_text_strategy,
'has_n_lines': not_implemented_strategy,
'has_n_columns': not_implemented_strategy,

# not Implemented
'ftype': ftype_strategy,
'not_has_text': not_implemented_strategy,
'has_text_matching': not_implemented_strategy,
'has_line': not_implemented_strategy,
'has_line_matching': not_implemented_strategy,
"""


def has_text(output_value: Any, kwargs: dict[str, Any]) -> bool:
    try:
        assert_has_text(output_value, kwargs['expected_value'], n=None)
    except AssertionError:
        return False
    return True

def has_n_columns(output_value: Any, kwargs: dict[str, Any]) -> bool:
    try:
        assert_has_n_columns(output_value, kwargs['expected_value'])
    except AssertionError:
        return False
    return True


def has_text_matching(output_value: Any, kwargs: dict[str, Any]) -> bool:
    try:
        assert_has_text_matching(output_value, kwargs['expected_value'])
    except AssertionError:
        return False
    return True







from abc import ABC, abstractmethod
from .token import Token, TokenType


### ORDERING ###

class TokenOrderingStrategy(ABC):
    @abstractmethod
    def order(self, token_list: list[Token]) -> list[Token]:
        """selects a token in the list according to some prioritisation"""
        ...

class FirstTokenOrderingStrategy(TokenOrderingStrategy):
    def order(self, token_list: list[Token]) -> list[Token]:
        """selects a token in the list according to some prioritisation"""
        token_list.sort(key=lambda x: x.start)
        return token_list

class LongestTokenOrderingStrategy(TokenOrderingStrategy):
    def order(self, token_list: list[Token]) -> list[Token]:
        """selects a token in the list according to some prioritisation"""
        token_list.sort(key=lambda x: x.end - x.start)
        return token_list

class PriorityTokenOrderingStrategy(TokenOrderingStrategy):
    def order(self, token_list: list[Token]) -> list[Token]:
        priorities: dict[TokenType, int] = {
            TokenType.FUNCTION_CALL: 0,
            TokenType.BACKTICK_SHELL_STATEMENT: 0,
            TokenType.SCRIPT: 1,
            TokenType.GX_KW_DYNAMIC: 1,
            TokenType.GX_KW_STATIC: 2,
            TokenType.KV_LINKER: 3,
            TokenType.GX_INPUT: 4,
            TokenType.GX_OUTPUT: 4,
            TokenType.ENV_VAR: 5,
            TokenType.STRING: 8,
            TokenType.INTEGER: 9,
            TokenType.FLOAT: 9,
            TokenType.LINUX_TEE: 10,
            TokenType.LINUX_REDIRECT: 10,
            TokenType.LINUX_STREAM_MERGE: 10,
            TokenType.EMPTY_STRING: 11,
            TokenType.UNKNOWN: 999,
        }
        token_list.sort(key=lambda x: priorities[x.ttype])
        return token_list





import os
import tarfile
from typing import Optional
from galaxy2janis.paths import USER_DATA_DIR, DOWNLOADED_WRAPPERS_DIR


class DownloadCache:
    """
    keeps track of the location of downloaded wrapper folders.
    DownloadCache.get() will return the local path to a tool xml if already downloaded
    DownloadCache.add() saves a tar as a download and notes its path. 
    """

    def get(self, query_repo: str, query_revision: str) -> Optional[str]:
        """returns the local file path for the tool xml if already downloaded or None"""
        path = f'{USER_DATA_DIR}/{DOWNLOADED_WRAPPERS_DIR}'
        for folder in self._load():
            repo, revision = folder.split('-', 1)
            if repo == query_repo and revision == query_revision:
                return f'{path}/{folder}'
        return None

    def add(self, tar: tarfile.TarFile) -> None:
        self._save(tar)

    def _save(self, tar: tarfile.TarFile) -> None:
        path = f'{USER_DATA_DIR}/{DOWNLOADED_WRAPPERS_DIR}'
        tar.extractall(path=path)

    def _load(self) -> set[str]:
        path = f'{USER_DATA_DIR}/{DOWNLOADED_WRAPPERS_DIR}'
        if not os.path.exists(path):
            return set()
        else:
            folders = os.listdir(path)
            folders = [f for f in folders if os.path.isdir(f'{path}/{f}')]
            return set(folders)





from typing import Any, Optional

from galaxy2janis.gx.command.components import InputComponent
from . import utils

from galaxy2janis.entities.workflow import (
    ConnectionInputValue, 
    WorkflowInputInputValue,
    StaticInputValue
)


def static(component: Optional[InputComponent], value: Any, default: bool=False) -> StaticInputValue:
    return StaticInputValue(
        component=component,
        value=value,
        _valtypestr=utils.select_input_value_type(component, value),
        default=default
    )

def connection(component: Optional[InputComponent], step_uuid: str, out_uuid: str) -> ConnectionInputValue:
    return ConnectionInputValue(
        component=component,
        step_uuid=step_uuid,
        out_uuid=out_uuid
    )

def workflow_input(component: Optional[InputComponent], input_uuid: str, is_runtime: bool=False) -> WorkflowInputInputValue:
    return WorkflowInputInputValue(
        component=component,
        input_uuid=input_uuid,
        is_runtime=is_runtime
    )








from __future__ import annotations
from typing import TYPE_CHECKING

from galaxy2janis.gx.command.components.inputs.InputComponent import InputComponent

if TYPE_CHECKING:
    from galaxy2janis.entities.workflow import WorkflowStep
    from galaxy2janis.entities.workflow import Workflow

from . import factory
from galaxy2janis import settings


def handle_tool_default_inputs(janis: Workflow) -> None:
    # sets tool input values as default
    for j_step in janis.steps:
        settings.tool.set(wrapper=j_step.metadata.wrapper)
        ingest_values_defaults(j_step)

def ingest_values_defaults(j_step: WorkflowStep) -> None:
    # tool components which don't yet appear in register
    for component in get_linkable_components(j_step):
        input_value = factory.static(component, component.default_value, default=True)
        j_step.inputs.add(input_value)

def get_linkable_components(j_step: WorkflowStep) -> list[InputComponent]:
    out: list[InputComponent] = []
    for component in j_step.tool.inputs:
        if not j_step.inputs.get(component.uuid):
            out.append(component)
    return out

        



from typing import Tuple

from galaxy2janis.entities.workflow import Workflow
from galaxy2janis.entities.workflow import WorkflowInput
from ..TextRender import TextRender
#from ..formatting import format_typestr

def note_snippet(commenter: str) -> str: # TODO
    return f"""{commenter} WORKFLOW INPUTS
{commenter} This file contains workflow inputs which need to be provided by the user.
{commenter} The names of workflow inputs (below) will appear in workflow.py where they are used. 

{commenter} NULL VALUES
{commenter} If a value should be left null, ensure the tool input is optional. 
{commenter} The engine will throw an error otherwise.  
"""

# def note_snippet(commenter: str) -> str: # TODO
#     return f"""{commenter} This file contains workflow inputs which need to be provided by the user.
# {commenter} Organised as follows: 
# {commenter}     1. input data for the workflow
# {commenter}     2. runtime values for each step

# {commenter} Null values must be replaced by the user to run the workflow. 
# {commenter} VIGNETTES
# """

class InputsText(TextRender):
    def __init__(self, entity: Workflow, file_format: str='yaml'):
        super().__init__()
        self.entity = entity
        self.file_format = file_format

    @property
    def imports(self) -> list[Tuple[str, str]]:
        raise NotImplementedError()

    def render(self) -> str:
        if self.file_format == 'yaml':
            return to_yaml(self.entity.inputs)
        else:
            raise RuntimeError('wrong format')



YAML_NONE = 'null'
YAML_COMMENTER = '#'

def to_yaml(inputs: list[WorkflowInput]) -> str:
    out_str: str = ''
    out_str += '\n'
    out_str += f'{note_snippet(commenter=YAML_COMMENTER)}'
    out_str += '\n'

    for winp in inputs:
        if winp.value is None:
            winp.value = YAML_NONE
        #out_str += f'{winp.tag}: {winp.value}  {YAML_COMMENTER} [{format_typestr(winp)}]\n'
        out_str += f'{winp.tag}: {winp.value}\n'

    return out_str


# future formats here





from typing import Callable
from .filters import (
    flatten_multiline_strings,
    translate_variable_markers,
    standardise_variable_format,
    simplify_sh_constructs,
    simplify_galaxy_dynamic_vars,
    remove_cheetah_comments,
    replace_function_calls,
    replace_backticks,
    remove_empty_quotes,
    interpret_raw
)



def simplify_cmd(text: str, output: str) -> str:
    simplifier_map = {
        'test': TestSimplifier(),
        'xml': XMLSimplifier(),
        'cheetah': PartialCheetahEvalSimplifier(),
    }
    simplifier = simplifier_map[output]
    return simplifier.simplify(text)



# classes 

class CommandSimplifier:
    filters: list[Callable[[str], str]] = []

    def simplify(self, cmdstr: str) -> str:
        return self.map_filters(cmdstr)

    def map_filters(self, cmdstr: str) -> str:
        for filter_func in self.filters:
            cmdstr = filter_func(cmdstr)
        return cmdstr


class PartialCheetahEvalSimplifier(CommandSimplifier):
    filters: list[Callable[[str], str]] = [
        remove_cheetah_comments,
        simplify_galaxy_dynamic_vars,
    ]


class TestSimplifier(CommandSimplifier):
    filters: list[Callable[[str], str]] = [
        translate_variable_markers,
        standardise_variable_format,
        simplify_galaxy_dynamic_vars,
        simplify_sh_constructs,
        replace_backticks,
        remove_empty_quotes,
    ]


class XMLSimplifier(CommandSimplifier):
    filters: list[Callable[[str], str]] = [
        remove_cheetah_comments,
        flatten_multiline_strings,
        replace_function_calls,
        replace_backticks,
        standardise_variable_format,  # ?
        simplify_sh_constructs,
        simplify_galaxy_dynamic_vars,
        remove_empty_quotes,
        interpret_raw
    ]






import argparse
from typing import Optional


useage_str = '''
gxtool2janis command [OPTIONS]

Commands:
   tool         Parse galaxy tool (.xml format)
   workflow     Parse galaxy workflow (.ga format)

'''

class CLIparser:
    def __init__(self, argv: list[str]):
        parser = argparse.ArgumentParser(
            description='gxtool2janis.py',
            usage=useage_str
        )
        parser.add_argument('command', help='Command')
        args = parser.parse_args(argv[1:2])
        if not hasattr(self, args.command):
            print(f'Unrecognized command {argv[1]}')
            parser.print_help()
            exit(1)
        # dispatch pattern (calls the function with same name as args.command)
        self.command = args.command
        getattr(self, args.command)(argv)

    def tool(self, argv: list[str]):
        parser = argparse.ArgumentParser(
            description='Parse single galaxy tool'
        )
        parser.add_argument("infile",
                            help="path to tool.xml file to parse.", 
                            type=str,
                            )
        parser.add_argument("-r",
                            "--remote", 
                            help="comma separated list of [owner],[repo],[toolid],[revision] if downloading wrapper from toolshed. example: devteam,picard,picard_MarkDuplicates,29:1aac2a13842a",
                            type=str
                            )
        parser.add_argument("-o",
                            "--outdir",
                            help="output folder to place translation", 
                            type=str,
                            )
        args = parser.parse_args(argv[2:])
        out: dict[str, Optional[str]] = args.__dict__
        out['command'] = self.command
        self.args = out

    def workflow(self, argv: list[str]):
        parser = argparse.ArgumentParser(
            description='') 
        parser.add_argument("infile", 
                            help="path to workflow.ga file to parse.", 
                            type=str)
        parser.add_argument("-o",
                            "--outdir",
                            help="output folder to place translation", 
                            type=str,
                            )
        parser.add_argument("--dev-partial-eval", 
                            help="turn off partial cheetah evaluation when identifying tool values",
                            default=False,
                            action='store_true')
        args = parser.parse_args(argv[2:])
        out: dict[str, Optional[str]] = args.__dict__
        out['command'] = self.command
        self.args = out







from typing import Any


def sanitise_default_value(value: Any) -> Any:
    value = _sanitise_env_var(value)
    return value

def _sanitise_env_var(value: Any) -> Any:
    if isinstance(value, str) and value.startswith('$'):
        value = None
    return value





from galaxy2janis.entities.workflow import Workflow
from galaxy2janis.entities.workflow.step.inputs import (
    InputValue, 
)


def update_components_datatype(value: InputValue, janis: Workflow) -> None:
    pass






from dataclasses import dataclass
from typing import Optional, Tuple

from galaxy2janis.gx.command.components import OutputComponent
from galaxy2janis.gx.command.components import InputOutput
from galaxy2janis.gx.command.components import RedirectOutput
from galaxy2janis.gx.command.components import WildcardOutput

from ..TextRender import TextRender
from .. import formatting
from .. import ordering

from galaxy2janis import tags  # remove if possible


def format_selector_str(output: OutputComponent) -> Optional[str]:
    match output: 
        case RedirectOutput():
            return None
        case InputOutput():
            input_comp_uuid = output.input_component.uuid
            input_comp_tag = tags.get(input_comp_uuid)
            return f'InputSelector("{input_comp_tag}")'
        case WildcardOutput():
            pattern = 'UNKNOWN'
            if hasattr(output.gxparam, 'from_work_dir') and output.gxparam.from_work_dir:
                pattern = output.gxparam.from_work_dir
            elif output.gxparam.discover_pattern:
                pattern = output.gxparam.discover_pattern
            return f'WildcardSelector(r"{pattern}")'
        case _:
            pass


@dataclass
class ToolOutputText(TextRender):
    def __init__(self, entity: OutputComponent):
        super().__init__()
        self.entity = entity

    @property
    def imports(self) -> list[Tuple[str, str]]:
        jtype = self.entity.datatype
        imports: list[Tuple[str, str]] = []
        imports.append((jtype.import_path, jtype.classname))

        # TODO opportunity for decorator # Stdout class import
        if isinstance(self.entity, RedirectOutput):
            imports.append(('janis_core', 'Stdout'))

        # TODO opportunity for decorator # Selector class import
        selector_str = format_selector_str(self.entity)
        if selector_str:
            selector = selector_str.split('(', 1)[0]
            imports.append(('janis_core', selector))
        
        # TODO opportunity for decorator # Array class import
        if self.entity.array:
            imports.append(('janis_core', 'Array'))

        # TODO opportunity for decorator
        imports = list(set(imports))
        return ordering.order_imports(imports)

    def render(self) -> str:
        e = self.entity
        selector_str = format_selector_str(e)
        doc = formatting.format_docstring(e)
        datatype_str = formatting.format_typestr(e)
        if isinstance(e, RedirectOutput):
            datatype_str = f'Stdout'
        out_str: str = ''
        out_str += '\tToolOutput(\n'
        out_str += f"\t\t'{tags.get(e.uuid)}',\n"
        out_str += f"\t\t{datatype_str},\n"
        if selector_str:
            out_str += f"\t\tselector={selector_str},\n" 
        out_str += f'\t\tdoc="{doc}",\n' if doc else ''
        out_str += '\t)'
        return out_str

    





from galaxy2janis.logs import logging
from copy import deepcopy
from typing import Any

from galaxy.tools.parameters.basic import ToolParameter, HiddenToolParameter
from galaxy.tools.parameters.grouping import Conditional, ConditionalWhen, Section, Repeat

XmlNode = ToolParameter | Conditional | ConditionalWhen | Section | Repeat


class ParamFlattener:
    def __init__(self, inputs: dict[str, Any]):
        self.inputs = inputs
        self.flat_params: list[ToolParameter] = [] 

    def flatten(self) -> list[ToolParameter]:
        for node in self.inputs.values():
            self.explore_node(node, [])
        return self.flat_params

    def explore_node(self, node: XmlNode, heirarchy: list[str]) -> None:
        heirarchy = deepcopy(heirarchy)
        match node:
            case HiddenToolParameter():
                pass
            case ToolParameter(): # leaf node
                self.flatten_param(node, heirarchy)
            case Conditional(): 
                heirarchy.append(node.name)
                self.flatten_param(node.test_param, heirarchy)
                for child in node.cases:
                    self.explore_node(child, heirarchy)
            case ConditionalWhen():
                for child in node.inputs.values():
                    self.explore_node(child, heirarchy)
            case Section():
                heirarchy.append(node.name)
                for child in node.inputs.values():
                    self.explore_node(child, heirarchy)
            case Repeat():
                logging.has_repeat()
            case _:
                raise NotImplementedError()
    
    def flatten_param(self, param: ToolParameter, heirarchy: list[str]) -> None:
        prefix = ".".join(heirarchy) + '.' if len(heirarchy) > 0 else ''
        flat_name = f'{prefix}{param.name}'
        param.flat_name = flat_name
        self.flat_params.append(param)







from typing import Optional

from ...gxtool.param.Param import Param

from .inputs.Flag import Flag
from .inputs.Option import Option
from .inputs.Positional import Positional
        
from .outputs.RedirectOutput import RedirectOutput
from .outputs.InputOutput import InputOutput
from .outputs.WildcardOutput import WildcardOutput
from .CommandComponent import CommandComponent


def positional(value: Optional[str]=None, gxparam: Optional[Param]=None) -> Positional:
    component = Positional()
    component.gxparam = gxparam
    if value:
        component.values.add(value)
    return component

def flag(prefix: str, gxparam: Optional[Param]=None) -> Flag:
    component = Flag(prefix=prefix)
    component.gxparam = gxparam
    return component

def option( 
    prefix: str,
    gxparam: Optional[Param]=None,
    delim: Optional[str]=None,
    values: Optional[list[str]]=None
) -> Option:
    component = Option(prefix=prefix)
    component.gxparam = gxparam
    if delim:
        component.delim = delim
    if values:
        for val in values:
            component.values.add(val)
    return component

def redirect_output(redirect: str, filepath: str, gxparam: Optional[Param]=None) -> RedirectOutput:
    output = RedirectOutput(redirect, filepath)
    output.gxparam = gxparam
    return output

def input_output(input_component: CommandComponent) -> InputOutput:
    return InputOutput(input_component)

def wildcard_output(gxparam: Param) -> WildcardOutput:
    output = WildcardOutput()
    output.gxparam = gxparam
    return output

def uncertain_output(gxparam: Param) -> WildcardOutput:
    output = WildcardOutput()
    output.gxparam = gxparam
    output.verified = False
    return output




from typing import Tuple

from galaxy2janis.entities.workflow import Workflow
from ..TextRender import TextRender


class ConfigText(TextRender):
    def __init__(self, entity: Workflow):
        super().__init__()
        self.entity = entity

    @property
    def imports(self) -> list[Tuple[str, str]]:
        raise NotImplementedError()

    def render(self) -> str:
        raise NotImplementedError()
    







import json
from typing import Any


def expand_tool_state(step: dict[str, Any]) -> dict[str, Any]:
    expander = ToolStateExpander(step)
    return expander.load()


class ToolStateExpander:
    def __init__(self, step: dict[str, Any]):
        self.step = step
        self.tool_state: dict[str, Any] = {}
        #self.tool_state = json.loads(step['tool_state'])

    def load(self) -> dict[str, Any]:
        self.tool_state = self.explore_node(self.step['tool_state'])
        return self.tool_state
    
    def explore_node(self, node: Any) -> Any:
        if self.is_quoted(node) or self.is_brackets(node):
            node = json.loads(node)
            if isinstance(node, dict):
                for key, val in node.items():
                    node[key] = self.explore_node(val) # type: ignore
        return node

    def is_quoted(self, node: Any) -> bool:
        if isinstance(node, str):
            if node.startswith('"') and node.endswith('"'): 
                return True
            elif node.startswith("'") and node.endswith("'"): 
                return True
        return False
    
    def is_brackets(self, node: Any) -> bool:
        if isinstance(node, str):
            if node.startswith('{') and node.endswith('}'): 
                return True
        return False

    






from abc import ABC, abstractmethod
import requests
import tarfile
import shutil
import os
from dataclasses import dataclass
from typing import Any, Optional
from xml.etree import ElementTree as et

from galaxy.tool_util.parser import get_tool_source
from galaxy2janis.gx.wrappers import Wrapper
import gx.wrappers.scraping.utils as utils
from wrappers.WrapperCache import WrapperCache


THREADS = 10


def scrape_wrappers() -> None:
    scraper = WrapperScraper()
    scraper.scrape()


class ScrapeStrategy(ABC):
    def __init__(self, owner: str, repo: str, revision: str):
        self.owner = owner
        self.repo = repo
        self.revision = revision

    @abstractmethod
    def scrape(self) -> None:
        raise NotImplementedError()


@dataclass
class Revision:
    owner: str
    repo: str
    revision: str
    date_created: str
    download_folder: str = 'temp'

    @property
    def url(self) -> str:
        return f'https://toolshed.g2.bx.psu.edu/repos/{self.owner}/{self.repo}/archive/{self.revision}.tar.gz'

    def get_tar(self) -> tarfile.TarFile:
        response = requests.get(self.url, stream=True)
        tar = tarfile.open(fileobj=response.raw, mode='r:gz')
        tar.extractall(path=self.download_folder)
        return tar

    def get_wrappers(self) -> list[Wrapper]:
        tar = self.get_tar()
        folder = os.path.commonprefix(tar.getnames())
        xmlpaths = [f'{self.download_folder}/{x}' for x in tar.getnames() if x.endswith('.xml')]
        wrappers = self.extract_wrapper_info(xmlpaths)
        shutil.rmtree(f'{self.download_folder}/{folder}')
        return wrappers

    def extract_wrapper_info(self, xmlpaths: list[str]) -> list[Wrapper]:
        out: list[Wrapper] = []
        for path in xmlpaths:
            info = self.get_info(path)
            if info:
                info['owner'] = self.owner
                info['repo'] = self.repo
                info['revision'] = self.revision
                info['date_created'] = self.date_created
                info['requirements'] = []
                raise NotImplementedError()
                out.append(Wrapper(info))
        return out
    
    def get_info(self, path: str) -> Optional[dict[str, Any]]:
        tree = et.parse(path)
        root = tree.getroot()
        if root.tag == 'tool':
            tool_source = get_tool_source(path)  # TODO macro paths????
            return {
                'tool_id': tool_source.parse_id(),  # type: ignore
                'tool_build': tool_source.parse_version()  # type: ignore
            }
        return None



class WrapperScraper:

    def __init__(self):
        self.cache = WrapperCache()
        self.total_count: int = 0
        self.scraped_count: int = 0

    @property
    def percentage_complete(self) -> float:
        return (self.scraped_count / self.total_count) * 100

    def scrape(self) -> None:
        revisions = self.get_scrapable_revisions()
        self.total_count = len(revisions)
        print(f'revisions to scrape: {self.total_count}')
        for revision in revisions:
            try:
                self.scraped_count += 1
                print(f'\r{self.percentage_complete:0.1f}%', end='')
                wrappers = revision.get_wrappers()
                for wrapper in wrappers:
                    self.cache.add(wrapper)
            except Exception as e:
                print(e)

    def get_scrapable_revisions(self) -> list[Revision]:
        revision_data = utils.load_data(utils.REVISION_DATA_PATH)
        out: list[Revision] = []
        for owner, repo_data in revision_data.items():
            for repo, revision_list in repo_data.items():
                for revision in revision_list:
                    if not self.cache.get(revision=revision['revision']):
                        new_revision = Revision(
                            owner=owner,
                            repo=repo,
                            revision=revision['revision'],
                            date_created=revision['date_created'],
                        )
                        out.append(new_revision)
        return out






def split_lines(cmdstr: str) -> list[str]:
    lines = cmdstr.split('\n')
    lines = [ln.strip() for ln in lines]
    lines = [ln for ln in lines if ln != '']
    return lines 

def split_lines_blanklines(cmdstr: str) -> list[str]:
    lines = cmdstr.split('\n')
    lines = [ln.strip() for ln in lines]
    return lines 

def split_lines_whitespace(cmdstr: str) -> list[str]:
    return cmdstr.split('\n')

def join_lines(cmdlines: list[str]) -> str:
    return '\n'.join(cmdlines)





from typing import Any

from ..load import load_xmltool
from .simplification.aliases import resolve_aliases
from .cheetah.evaluation import sectional_evaluate
from .simplification.simplify import simplify_cmd


def load_command() -> str:
    xmltool = load_xmltool()
    text = simplify_cmd(xmltool.raw_command, 'xml')
    text = resolve_aliases(text)
    return text

def load_partial_cheetah_command(inputs_dict: dict[str, Any]) -> str:
    xmltool = load_xmltool()
    text = xmltool.raw_command
    text = simplify_cmd(text, 'cheetah')
    text = sectional_evaluate(text, inputs=inputs_dict)
    text = resolve_aliases(text)
    return text







from abc import ABC, abstractmethod
from typing import Any

from ..Container import Container



class Fetcher(ABC):
    """
    searched for the most recent container build which matches the name and version
    can just format a container requirement, or use an API to search online
    individual builds not supported, except in the case of of 'container' requirements as it just directly grabs the url from the requirement text 
    """

    @abstractmethod
    def fetch(self, requirement: Any) -> list[Container]:
        """finds relevant biocontainers for a given tool requirement"""
        ...






# oi scroll to bottom 

from .JanisDatatype import JanisDatatype

file_t = JanisDatatype(
    format='file',
    source='janis',
    classname='File',
    extensions=None,
    import_path='janis_core.types.common_data_types'
)
string_t = JanisDatatype(
    format='string',
    source='janis',
    classname='String',
    extensions=None,
    import_path='janis_core.types.common_data_types'
)
directory_t = JanisDatatype(
    format='directory',
    source='janis',
    classname='Directory',
    extensions=None,
    import_path='janis_core.types.common_data_types'
)
int_t = JanisDatatype(
    format='integer',
    source='janis',
    classname='Int',
    extensions=None,
    import_path='janis_core.types.common_data_types'
)
float_t = JanisDatatype(
    format='float',
    source='janis',
    classname='Float',
    extensions=None,
    import_path='janis_core.types.common_data_types'
)
bool_t = JanisDatatype(
    format='boolean',
    source='janis',
    classname='Boolean',
    extensions=None,
    import_path='janis_core.types.common_data_types'
)

DEFAULT_DATATYPE = file_t

CORE_DATATYPES = set([
    file_t.classname,
    string_t.classname,
    directory_t.classname,
    int_t.classname,
    float_t.classname,
    bool_t.classname
])






from __future__ import annotations
from typing import Optional

from galaxy2janis.entities.tool import Tool
from uuid import uuid4
from galaxy2janis import tags

from ..registers.StepInputRegister import StepInputRegister
from ..registers.StepOutputRegister import StepOutputRegister
from .metadata import StepMetadata


class WorkflowStep:
    """represents a galaxy tool step"""

    def __init__(self, metadata: StepMetadata):
        self.metadata = metadata
        self.uuid: str = str(uuid4())
        self.inputs: StepInputRegister = StepInputRegister()
        self.outputs: StepOutputRegister = StepOutputRegister()
        self.preprocessing: Optional[str] = None
        self.postprocessing: Optional[str] = None
        self._tool: Optional[Tool] = None

    @property
    def name(self) -> str:
        return self.metadata.wrapper.tool_id
    
    @property
    def tag(self) -> str:
        return tags.get(self.uuid)
    
    @property
    def docstring(self) -> Optional[str]:
        return self.metadata.label

    @property
    def tool(self) -> Tool:
        if self._tool:
            return self._tool
        raise RuntimeError()

    def set_tool(self, tool: Tool) -> None:
        self._tool = tool

    def __repr__(self) -> str:
        return f'(WorkflowStep) step{self.metadata.step_id} - {self.name}'







from galaxy2janis.entities.workflow.input import WorkflowInput
from galaxy2janis.entities.workflow.step.step import WorkflowStep
from galaxy2janis.entities.workflow.workflow import Workflow


def handle_scattering(janis: Workflow) -> None:
    """
    Identify the inputs in each step which need to be scattered.
    Workflow inputs which are arrays trigger any downstream steps
    to be scattered. 
    Still unsure if array outputs should also trigger scattering.
    """
    # get workflow inputs which are arrays
    for winp in janis.inputs:
        if winp.array: 
            # scatter direct child steps, then recursively scatter
            # child steps which connect to affected steps
            children = janis.get_input_children(winp.uuid)
            for child in children:
                update_scatter_inputs(child, parent=winp)
                recursive_scatter(child, janis)

def recursive_scatter(step: WorkflowStep, janis: Workflow) -> None:
    """
    Identify inputs which will be scattered on.
    Then do for child steps.
    """
    children = janis.get_step_children(step.uuid)
    for child in children:
        update_scatter_inputs(child, parent=step)
        recursive_scatter(child, janis)

def update_scatter_inputs(step: WorkflowStep, parent: WorkflowInput | WorkflowStep):
    match parent:
        case WorkflowInput():
            invalues = step.inputs.workflow_inputs
            for val in invalues:
                if val.input_uuid == parent.uuid:
                    val.scatter = True
        case WorkflowStep():
            invalues = step.inputs.connections
            for val in invalues:
                if val.step_uuid == parent.uuid:
                    val.scatter = True
        case _:
            raise RuntimeError()






from typing import Optional
from galaxy2janis.gx.gxtool import XMLToolDefinition
from galaxy2janis.gx.command import Command

# this module imports
from .Tool import Tool
from .ToolFactory import ToolFactory


def gen_tool(xmltool: XMLToolDefinition, command: Command, container: Optional[str]) -> Tool:
    factory = ToolFactory(xmltool, command, container)
    tool = factory.create()
    return tool






from .workflow import Workflow
from .metadata import WorkflowMetadata
from .input import WorkflowInput

from .registers.StepOutputRegister import StepOutputRegister
from .registers.StepInputRegister import StepInputRegister

from .step.inputs import ( 
    InputValue,
    ConnectionInputValue,
    WorkflowInputInputValue,
    StaticInputValue
)
    
from .step.step import WorkflowStep
from .step.metadata import StepMetadata
from .step.outputs import StepOutput








import shlex


def split_lines(cmdstr: str) -> list[str]:
    sh = shlex.shlex(cmdstr)
    sh.commenters = ''
    sh.whitespace = '\n'
    sh.whitespace_split = True
    lines = list(sh)
    lines = [ln.strip() for ln in lines]
    lines = [ln for ln in lines if ln != '']
    return lines 

def join_lines(lines: list[str]) -> str:
    return '\n'.join(lines)

def split_to_words(line: str) -> list[str]:
    return shlex.split(line, posix=False)







from ..step.outputs import StepOutput


class StepOutputRegister:
    def __init__(self):
        self.register: list[StepOutput] = []

    def add(self, step_output: StepOutput) -> None:
        self.register.append(step_output)

    def list(self) -> list[StepOutput]:
        return self.register
        







from galaxy2janis.entities.tool import Tool
from galaxy2janis.entities.workflow import Workflow
from galaxy2janis.entities.workflow import WorkflowStep
from .Format import Format


class SubworkflowFormat(Format):
    """
    renders output pages in different manner as appears in original workflow.
    steps using tools which have pre/post process are broken into 3 step subworkflows. 
    """

    def workflow(self, workflow: Workflow) -> str:
        raise NotImplementedError()
    
    def subworkflow(self, workflow: Workflow) -> str:
        raise NotImplementedError()
    
    def step(self, step: WorkflowStep) -> str:
        raise NotImplementedError()

    def tool(self, tool: Tool) -> str:
        raise NotImplementedError()

    def script(self) -> str:
        raise NotImplementedError()
    
    def inputs(self, workflow: Workflow) -> str:
        raise NotImplementedError()
    
    def config(self, workflow: Workflow) -> str:
        raise NotImplementedError()







from abc import ABC, abstractmethod

from galaxy2janis.entities.tool import Tool
from galaxy2janis.entities.workflow import Workflow
from galaxy2janis.entities.workflow import WorkflowStep


class Format(ABC):
    """
    controls the format of pages user will see.
    example 1: controls whether the workflow steps will be 
    broken into subworkflows or not. 
    example 2?: controls which workflows will have an inputs dict file created? 
    """

    @abstractmethod
    def workflow(self, workflow: Workflow) -> str:
        """returns the workflow page"""
        ...
    
    @abstractmethod
    def subworkflow(self, workflow: Workflow) -> str:
        """returns a subworkflow"""
        ...
    
    @abstractmethod
    def step(self, step: WorkflowStep) -> str:
        """writes a step"""
        ...

    @abstractmethod
    def tool(self, tool: Tool) -> str:
        """writes a tool"""
        ...

    @abstractmethod
    def script(self) -> str:
        """writes a script used with a tool"""
        ...
    
    @abstractmethod
    def inputs(self, workflow: Workflow) -> str:
        """writes the workflow inputs file"""
        ...
    
    @abstractmethod
    def config(self, workflow: Workflow) -> str:
        """writes the workflow config file"""
        ...







import regex as re

from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional

from galaxy2janis.gx.gxtool.param import Param
from galaxy2janis.gx.command.cmdstr import constructs

class TokenType(Enum):
    FUNCTION_CALL   = auto()
    SCRIPT          = auto()
    GX_INPUT        = auto()
    GX_OUTPUT       = auto()
    GX_KW_DYNAMIC   = auto()
    GX_KW_STATIC    = auto()
    ENV_VAR         = auto()
    KV_LINKER       = auto()
    FORCED_PREFIX   = auto()
    BACKTICK_SHELL_STATEMENT = auto()
    STRING          = auto()
    INTEGER         = auto()
    FLOAT           = auto()
    LINUX_TEE       = auto()
    LINUX_REDIRECT  = auto()
    LINUX_STREAM_MERGE = auto()
    START_STATEMENT = auto()
    EXCISION        = auto()
    END_STATEMENT   = auto()
    EMPTY_STRING    = auto()
    UNKNOWN         = auto()


@dataclass
class Token:
    match: re.Match[str]
    ttype: TokenType

    quoted: bool=False
    gxparam: Optional[Param]=None
    position: Optional[int] = None
    in_conditional: bool = False
    in_loop: bool = False
    construct: Optional[constructs.Construct] = None

    @property
    def text(self) -> str:
        return self.match[0] # type: ignore
    
    @property
    def start(self) -> int:
        return self.match.start()

    @property
    def end(self) -> int:
        return self.match.end()

    def __repr__(self) -> str:
        return f'Token: {self.text}, gxparam: {self.gxparam}'

